#!/usr/bin/python
# service_broadcasting.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_broadcasting.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_broadcasting
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return BroadcastingService()


class BroadcastingService(LocalService):

    service_name = 'service_broadcasting'
    config_path = 'services/broadcasting/enabled'

    scope = []  # set to [idurl1,idurl2,...] to receive only messages broadcasted from certain nodes

    # TODO: need to be able dynamically add/remove scope after start of the service
    # for now let's just listen for all broadcast messages (global scope)

    def dependent_on(self):
        return [
            'service_p2p_hookups',
            'service_nodes_lookup',
        ]

    def installed(self):
        # TODO: to be continue...
        return False

    def start(self):
        from twisted.internet.defer import Deferred
        from broadcast import broadcasters_finder
        from broadcast import broadcaster_node
        from broadcast import broadcast_listener
        from broadcast import broadcast_service
        from bitdust.main.config import conf
        from bitdust.main import settings
        self.starting_deferred = Deferred()
        broadcasters_finder.A('init')
        if settings.enableBroadcastRouting():
            broadcaster_node.A('init', broadcast_service.on_incoming_broadcast_message)
            broadcaster_node.A().addStateChangedCallback(self._on_broadcaster_node_switched)
        else:
            broadcast_listener.A('init', broadcast_service.on_incoming_broadcast_message)
            broadcast_listener.A().addStateChangedCallback(self._on_broadcast_listener_switched)
            broadcast_listener.A('connect', self.scope)
        conf().addConfigNotifier('services/broadcasting/routing-enabled', self._on_broadcast_routing_enabled_disabled)
        return self.starting_deferred

    def stop(self):
        from broadcast import broadcaster_node
        from broadcast import broadcasters_finder
        from broadcast import broadcast_listener
        from bitdust.main.config import conf
        broadcasters_finder.A('shutdown')
        if broadcaster_node.A() is not None:
            broadcaster_node.A().removeStateChangedCallback(self._on_broadcaster_node_switched)
            broadcaster_node.A('shutdown')
        if broadcast_listener.A() is not None:
            broadcast_listener.A().removeStateChangedCallback(self._on_broadcast_listener_switched)
            broadcast_listener.A('shutdown')
        conf().removeConfigNotifier('services/broadcasting/routing-enabled')
        return True

    def request(self, json_payload, newpacket, info):
        from bitdust.logs import lg
        from bitdust.p2p import p2p_service
        from bitdust.main import settings
        # words = newpacket.Payload.split(' ')
        try:
            mode = json_payload['action']
        except:
            lg.exc()
            return p2p_service.SendFail(newpacket, 'invalid payload')
        if mode != 'route' and mode != 'listen':
            lg.out(8, 'service_broadcasting.request DENIED, wrong mode provided : %s' % mode)
            return p2p_service.SendFail(newpacket, 'invalid request')
        if not settings.enableBroadcastRouting():
            lg.out(8, 'service_broadcasting.request DENIED, broadcast routing disabled')
            return p2p_service.SendFail(newpacket, 'broadcast routing disabled')
        from broadcast import broadcaster_node
        if not broadcaster_node.A():
            lg.out(8, 'service_broadcasting.request DENIED, broadcast routing disabled')
            return p2p_service.SendFail(newpacket, 'broadcast routing disabled')
        if broadcaster_node.A().state not in [
            'BROADCASTING',
            'OFFLINE',
            'BROADCASTERS?',
        ]:
            lg.out(8, 'service_broadcasting.request DENIED, current state is : %s' % broadcaster_node.A().state)
            return p2p_service.SendFail(newpacket, 'currently not broadcasting')
        if mode == 'route':
            broadcaster_node.A('new-broadcaster-connected', newpacket.OwnerID)
            lg.out(8, 'service_broadcasting.request ACCEPTED, mode: %s' % mode)
            return p2p_service.SendAck(newpacket, 'accepted')
        if mode == 'listen':
            # TODO: fix!!!
            # broadcaster_node.A().add_listener(newpacket.OwnerID, ' '.join(words[2:]))
            lg.out(8, 'service_broadcasting.request ACCEPTED, mode: %s' % mode)
            return p2p_service.SendAck(newpacket, 'accepted')
        return p2p_service.SendAck(newpacket, 'bad request')

    def health_check(self):
        from broadcast import broadcaster_node
        return broadcaster_node.A().state in [
            'BROADCASTING',
        ]

    def _on_broadcast_routing_enabled_disabled(self, path, value, oldvalue, result):
        from bitdust.logs import lg
        from broadcast import broadcaster_node
        from broadcast import broadcast_listener
        from broadcast import broadcast_service
        lg.out(2, 'service_broadcasting._on_broadcast_routing_enabled_disabled : %s->%s : %s' % (oldvalue, value, path))
        if not value:
            if broadcaster_node.A() is not None:
                broadcaster_node.A().removeStateChangedCallback(self._on_broadcaster_node_switched)
                broadcaster_node.A('shutdown')
            broadcast_listener.A('init', broadcast_service.on_incoming_broadcast_message)
            broadcast_listener.A().addStateChangedCallback(self._on_broadcast_listener_switched)
            broadcast_listener.A('connect', self.scope)
        else:
            if broadcast_listener.A() is not None:
                broadcast_listener.A().removeStateChangedCallback(self._on_broadcast_listener_switched)
                broadcast_listener.A('shutdown')
            broadcaster_node.A('init', broadcast_service.on_incoming_broadcast_message)
            broadcaster_node.A().addStateChangedCallback(self._on_broadcaster_node_switched)

    def _on_broadcast_listener_switched(self, oldstate, newstate, evt, *args, **kwargs):
        from bitdust.logs import lg
        from twisted.internet import reactor  # @UnresolvedImport
        from broadcast import broadcast_listener
        if self.starting_deferred:
            if newstate in [
                'LISTENING',
                'OFFLINE',
            ]:
                self.starting_deferred.callback(newstate)
                self.starting_deferred = None
        if newstate == 'OFFLINE':
            reactor.callLater(60, broadcast_listener.A, 'connect', self.scope)  # @UndefinedVariable
            lg.out(8, 'service_broadcasting._on_broadcast_listener_switched will try to connect again after 1 minute')

    def _on_broadcaster_node_switched(self, oldstate, newstate, evt, *args, **kwargs):
        from bitdust.logs import lg
        from twisted.internet import reactor  # @UnresolvedImport
        from broadcast import broadcaster_node
        if self.starting_deferred:
            if newstate in [
                'BROADCASTING',
            ] and newstate != oldstate:
                self.starting_deferred.callback(newstate)
                self.starting_deferred = None
            elif newstate in [
                'OFFLINE',
            ] and newstate != oldstate:
                self.starting_deferred.errback(Exception(newstate))
                self.starting_deferred = None


#         if newstate == 'OFFLINE' and oldstate != 'AT_STARTUP':
#             reactor.callLater(60, broadcaster_node.A, 'reconnect')  # @UndefinedVariable
#             lg.out(8, 'service_broadcasting._on_broadcaster_node_switched will try to reconnect again after 1 minute')

#     def cancel(self, json_payload, request, info):
#         from bitdust.logs import lg
#         from bitdust.p2p import p2p_service
#         words = request.Payload.split(' ')
#         try:
#             mode = words[1][:20]
#         except:
#             lg.exc()
#             return p2p_service.SendFail(request, 'wrong mode provided')
#         if mode == 'route' and False: # and not settings.getBroadcastRoutingEnabled():
#             # TODO check if this is enabled in settings
#             # so broadcaster_node should be existing already
#             lg.out(8, "service_broadcasting.request DENIED, broadcast routing disabled")
#             return p2p_service.SendFail(request, 'broadcast routing disabled')
#         from broadcast import broadcaster_node
#         if broadcaster_node.A().state not in ['BROADCASTING', ]:
#             lg.out(8, "service_broadcasting.request DENIED, current state is : %s" % broadcaster_node.A().state)
#             return p2p_service.SendFail(request, 'currently not broadcasting')
#         broadcaster_node.A('broadcaster-disconnected', request)
#         return p2p_service.SendAck(request, 'accepted')



#!/usr/bin/python
# udp_interface.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (udp_interface.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: udp_interface
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

import six

#------------------------------------------------------------------------------

import os
import sys

#------------------------------------------------------------------------------

try:
    pass
except:
    sys.exit('Error initializing twisted.internet.reactor in udp_interface.py')

from twisted.web import xmlrpc
from twisted.internet.defer import Deferred, succeed, fail

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import nameurl

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

_GateProxy = None

#------------------------------------------------------------------------------


def proxy():
    global _GateProxy
    return _GateProxy


#------------------------------------------------------------------------------


def idurl_to_id(idurl):
    proto, host, port, filename = nameurl.UrlParse(idurl)
    assert proto == 'http'
    user_id = filename.replace('.xml', '') + '@' + host
    if port and port not in [
        '80',
        80,
    ]:
        user_id += ':%s' % str(port)
    return user_id


def id_to_idurl(user_id):
    try:
        filename, host = user_id.split('@')
        filename += '.xml'
    except:
        return None
    return 'http://%s/%s' % (host, filename)


#------------------------------------------------------------------------------


class GateInterface():
    def init(self, xml_rpc_url_or_object):
        """
        """
        global _GateProxy
        if _Debug:
            lg.out(4, 'udp_interface.init %s' % xml_rpc_url_or_object)
        if isinstance(xml_rpc_url_or_object, six.string_types):
            _GateProxy = xmlrpc.Proxy(xml_rpc_url_or_object, allowNone=True)
        else:
            _GateProxy = xml_rpc_url_or_object
        _GateProxy.callRemote('transport_initialized', 'udp')
        return True

    def shutdown(self):
        """
        """
        from bitdust.transport.udp import udp_node
        global _GateProxy
        if _Debug:
            lg.out(4, 'udp_interface.shutdown')
        udp_node.Destroy()
        if _GateProxy:
            # del _GateProxy
            _GateProxy = None
        return succeed(True)

    def connect(self, options):
        """
        """
        from bitdust.transport.udp import udp_node
        if _Debug:
            lg.out(8, 'udp_interface.connect %s' % str(options))
        udp_node.A('go-online', options)
        return True

    def disconnect(self):
        """
        """
        from bitdust.transport.udp import udp_node
        if _Debug:
            lg.out(4, 'udp_interface.disconnect')
        udp_node.A('go-offline')
        return succeed(True)

    def build_contacts(self, id_obj):
        """
        """
        result = []
        result.append('udp://%s@%s' % (id_obj.getIDName().lower(), id_obj.getIDHost()))
        if _Debug:
            lg.out(4, 'udp_interface.build_contacts : %s' % str(result))
        return result

    def verify_contacts(self, id_obj):
        """
        """
        udp_contact = 'udp://%s@%s' % (id_obj.getIDName().lower(), id_obj.getIDHost())
        if id_obj.getContactIndex(contact=udp_contact) < 0:
            if _Debug:
                lg.out(4, 'udp_interface.verify_contacts returning False: udp contact not found or changed')
            return False
        if _Debug:
            lg.out(4, 'udp_interface.verify_contacts returning True')
        return True

    def send_file(self, remote_idurl, filename, host, description=''):
        """
        """
        from bitdust.transport.udp import udp_session
        from bitdust.transport.udp import udp_node
        # lg.out(20, 'udp_interface.send_file %s %s %s' % (filename, host, description))
        result_defer = Deferred()
        #        if udp_node.A().state not in ['LISTEN', 'DHT_READ',]:
        #            result_defer.callback(False)
        #            lg.out(4, 'udp_interface.send_file WARNING udp_node state is %s' % udp_node.A().state)
        #            return result_defer
        active_sessions = udp_session.get_by_peer_id(host)
        if active_sessions:
            if description.startswith('Identity') or description.startswith('Ack'):
                active_sessions[0].file_queue.insert_outbox_file(filename, description, result_defer, keep_alive=True)
            else:
                active_sessions[0].file_queue.append_outbox_file(filename, description, result_defer, keep_alive=True)
        else:
            udp_session.add_pending_outbox_file(filename, host, description, result_defer, keep_alive=True)
            udp_node.A('connect', host)
        return result_defer

    def send_file_single(self, remote_idurl, filename, host, description=''):
        """
        """
        return self.send_file(self, remote_idurl, filename, host, description, keep_alive=False)

    def send_keep_alive(self, host):
        """
        """
        from bitdust.transport.udp import udp_session
        for sess in udp_session.sessions_by_peer_id().get(host, []):
            sess.automat('send-keep-alive')

    def connect_to_host(self, host=None, idurl=None):
        """
        """
        from bitdust.transport.udp import udp_node
        if not host:
            host = idurl_to_id(idurl)
        if _Debug:
            lg.out(12, 'udp_interface.connect %s' % host)
        udp_node.A('connect', host)

    def disconnect_from_host(self, host):
        """
        """

    def cancel_outbox_file(self, host, filename):
        """
        """
        from bitdust.transport.udp import udp_session
        ok = False
        for sess in udp_session.sessions().values():
            if sess.peer_id != host:
                continue
            i = 0
            while i < len(sess.file_queue.outboxQueue):
                fn, descr, result_defer, keep_alive = sess.file_queue.outboxQueue[i]
                if fn == filename:
                    if _Debug:
                        lg.out(14, 'udp_interface.cancel_outbox_file removed %s in %s' % (os.path.basename(fn), sess))
                    sess.file_queue.outboxQueue.pop(i)
                    ok = True
                else:
                    i += 1
        udp_session.remove_pending_outbox_file(host, filename)
        #        for fn, descr, result_defer, single in sess.file_queue.outboxQueue:
        #            if fn == filename and sess.peer_id == host:
        #                lg.out(6, 'udp_interface.cancel_outbox_file    host=%s  want to close session' % host)
        #                sess.automat('shutdown')
        #                return True
        return ok

    def cancel_file_sending(self, transferID):
        """
        """
        from bitdust.transport.udp import udp_session
        for sess in udp_session.sessions().values():
            for out_file in sess.file_queue.outboxFiles.values():
                if out_file.transfer_id and out_file.transfer_id == transferID:
                    out_file.cancel()
                    return True
        return False

    def cancel_file_receiving(self, transferID):
        """
        """
        # at the moment for UDP transport we can not stop particular file transfer
        # we can only close the whole session which is not we really want
        #         for sess in udp_session.sessions().values():
        #             for in_file in sess.file_queue.inboxFiles.values():
        #                 if in_file.transfer_id and in_file.transfer_id == transferID:
        #                     if _Debug:
        #                         lg.out(6, 'udp_interface.cancel_file_receiving transferID=%s   want to close session' % transferID)
        #                     sess.automat('shutdown')
        #                     return True
        #         return False
        return False

    def list_sessions(self):
        """
        """
        from bitdust.transport.udp import udp_session
        return list(udp_session.sessions().values())

    def list_streams(self, sorted_by_time=True):
        """
        """
        from bitdust.transport.udp import udp_stream
        result = []
        for stream in udp_stream.streams().values():
            result.append(stream.consumer)
        if sorted_by_time:
            result.sort(key=lambda stream: stream.started)
        return result

    def find_session(self, host=None, idurl=None):
        """
        """
        from bitdust.transport.udp import udp_session
        if idurl:
            return udp_session.sessions_by_peer_id().get(idurl_to_id(idurl), [])
        return udp_session.sessions_by_peer_id().get(host, [])

    def find_stream(self, stream_id=None, transfer_id=None):
        """
        """
        from bitdust.transport.udp import udp_stream
        for stream in udp_stream.streams().values():
            if stream_id and stream_id == stream.consumer.stream_id:
                return stream.consumer
            if transfer_id and transfer_id == stream.consumer.transfer_id:
                return stream.consumer
        return None


#------------------------------------------------------------------------------


def proxy_errback(x):
    if _Debug:
        lg.out(6, 'udp_interface.proxy_errback ERROR %s' % x)
    return None


#------------------------------------------------------------------------------


def interface_transport_initialized():
    if proxy():
        return proxy().callRemote('transport_initialized', 'udp').addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_receiving_started(host, new_options={}):
    if proxy():
        return proxy().callRemote('receiving_started', 'udp', host, new_options).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_receiving_failed(error_code=None):
    if proxy():
        return proxy().callRemote('receiving_failed', 'udp', error_code).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_disconnected(result=None):
    if proxy():
        return proxy().callRemote('disconnected', 'udp', result).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    # return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)
    return succeed(result)


def interface_register_file_sending(host, receiver_idurl, filename, size, description=''):
    if proxy():
        return proxy().callRemote('register_file_sending', 'udp', host, receiver_idurl, filename, size, description).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_register_file_receiving(host, sender_idurl, filename, size):
    if proxy():
        return proxy().callRemote('register_file_receiving', 'udp', host, sender_idurl, filename, size).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_unregister_file_sending(transfer_id, status, bytes_sent, error_message=None):
    if proxy():
        return proxy().callRemote('unregister_file_sending', transfer_id, status, bytes_sent, error_message).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_unregister_file_receiving(transfer_id, status, bytes_received, error_message=None):
    if proxy():
        return proxy().callRemote('unregister_file_receiving', transfer_id, status, bytes_received, error_message).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_cancelled_file_sending(host, filename, size, description=None, error_message=None):
    if proxy():
        return proxy().callRemote('cancelled_file_sending', 'udp', host, filename, size, description, error_message).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)


def interface_cancelled_file_receiving(host, filename, size, error_message=None):
    if proxy():
        return proxy().callRemote('cancelled_file_receiving', 'udp', host, filename, size, error_message).addErrback(proxy_errback)
    lg.warn('transport_udp is not ready')
    return fail(Exception('transport_udp is not ready')).addErrback(proxy_errback)



#!/usr/bin/python
# message_database.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (message_database.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: message_database
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import json
import sqlite3

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime
from bitdust.lib import strng

from bitdust.main import settings
from bitdust.main import listeners

from bitdust.crypt import my_keys

from bitdust.access import group_participant

from bitdust.p2p import online_status

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_HistoryDB = None
_HistoryCursor = None

#------------------------------------------------------------------------------

MESSAGE_TYPES = {
    'message': 1,
    'private_message': 2,
    'group_message': 3,
    'personal_message': 4,
}

MESSAGE_TYPE_CODES = {
    1: 'message',
    2: 'private_message',
    3: 'group_message',
    4: 'personal_message',
}

#------------------------------------------------------------------------------


def init(filepath=None):
    global _HistoryDB
    global _HistoryCursor

    if not filepath:
        filepath = settings.ChatMessagesHistoryDatabaseFile()

    if _Debug:
        lg.args(_DebugLevel, filepath=filepath)

    sqlite3.register_adapter(dict, adapt_json)
    sqlite3.register_adapter(list, adapt_json)
    sqlite3.register_adapter(tuple, adapt_json)
    sqlite3.register_converter('JSON', convert_json)

    if not os.path.isfile(filepath):
        _HistoryDB = sqlite3.connect(filepath, timeout=1)
        _HistoryDB.execute('PRAGMA case_sensitive_like = 1;')
        _HistoryCursor = _HistoryDB.cursor()

        _HistoryCursor.execute(
            '''CREATE TABLE IF NOT EXISTS "history" (
            "sender_local_key_id" INTEGER,
            "sender_id" TEXT,
            "recipient_local_key_id" INTEGER,
            "recipient_id" TEXT,
            "direction" INTEGER,
            "payload_type" INTEGER,
            "payload_time" INTEGER,
            "payload_message_id" TEXT,
            "payload_body" JSON)''',
        )
        _HistoryCursor.execute('CREATE INDEX "sender local key id" on history(sender_local_key_id)')
        _HistoryCursor.execute('CREATE INDEX "recipient local key id" on history(recipient_local_key_id)')

        _HistoryCursor.execute(
            '''CREATE TABLE IF NOT EXISTS "conversations" (
            "conversation_id" TEXT,
            "payload_type" INTEGER,
            "started_time" INTEGER,
            "last_updated_time" INTEGER,
            "last_message_id" TEXT)''',
        )
        _HistoryCursor.execute('CREATE INDEX "conversation id" on conversations(conversation_id)')

        _HistoryCursor.execute('''CREATE TABLE IF NOT EXISTS "keys" (
            "key_id" TEXT,
            "local_key_id" INTEGER,
            "public_key" TEXT)''')
        _HistoryCursor.execute('CREATE INDEX "key id" on keys(key_id)')
        _HistoryCursor.execute('CREATE INDEX "local key id" on keys(local_key_id)')
        _HistoryCursor.execute('CREATE INDEX "public key" on keys(public_key)')

        _HistoryDB.commit()
        _HistoryDB.close()

    _HistoryDB = sqlite3.connect(filepath, timeout=1)
    _HistoryDB.text_factory = str
    _HistoryDB.execute('PRAGMA case_sensitive_like = 1;')
    _HistoryDB.commit()
    _HistoryCursor = _HistoryDB.cursor()


def shutdown():
    global _HistoryDB
    global _HistoryCursor
    if _Debug:
        lg.dbg(_DebugLevel, '')

    _HistoryDB.commit()
    _HistoryDB.close()
    _HistoryDB = None
    _HistoryCursor = None


#------------------------------------------------------------------------------


def db(instance='current'):
    global _HistoryDB
    return _HistoryDB


def cur(instance='current'):
    global _HistoryCursor
    return _HistoryCursor


#------------------------------------------------------------------------------


def adapt_json(data):
    return (json.dumps(data, sort_keys=True)).encode()


def convert_json(blob):
    return json.loads(blob.decode())


#------------------------------------------------------------------------------


def get_conversation_id(
    sender_local_key_id,
    recipient_local_key_id,
    payload_type,
):
    conversation_id = None
    if payload_type in [3, 4]:
        conversation_id = '{}&{}'.format(recipient_local_key_id, recipient_local_key_id)
    elif payload_type == 2:
        if recipient_local_key_id < sender_local_key_id:
            conversation_id = '{}&{}'.format(recipient_local_key_id, sender_local_key_id)
        else:
            conversation_id = '{}&{}'.format(sender_local_key_id, recipient_local_key_id)
    else:
        lg.err('unexpected message type: %r' % payload_type)
        return None
    return conversation_id


#------------------------------------------------------------------------------


def build_json_message(data, message_id, message_time=None, sender=None, recipient=None, message_type=None, direction=None, conversation_id=None):
    if not sender:
        sender = my_id.getGlobalID(key_alias='master')
    if not recipient:
        recipient = my_id.getGlobalID(key_alias='master')
    if direction is None:
        if message_type in ['private_message', None]:
            direction = 'out' if sender == my_id.getGlobalID(key_alias='master') else 'in'
        else:
            direction = 'in'
    else:
        direction = direction.replace('incoming', 'in').replace('outgoing', 'out')
    new_json = {
        'payload': {
            'msg_type': message_type or 'message',
            'message_id': strng.to_text(message_id),
            'time': message_time or utime.utcnow_to_sec1970(),
            'data': data,
        },
        'sender': {
            'glob_id': sender,
        },
        'recipient': {
            'glob_id': recipient,
        },
        'direction': direction,
        'conversation_id': conversation_id,
    }
    return new_json


def build_json_conversation(**record):
    conv = {
        'key_id': '',
        'label': '',
        'state': 'DISCONNECTED',
        'index': None,
        'id': None,
        'name': None,
        'events': None,
    }
    conv.update(record)
    if conv['type'] == 'private_message':
        local_key_id1, _, local_key_id2 = conv['conversation_id'].partition('&')
        try:
            local_key_id1 = int(local_key_id1)
            local_key_id2 = int(local_key_id2)
        except:
            lg.exc()
            return None
        usr1 = my_keys.get_local_key(local_key_id1)
        usr2 = my_keys.get_local_key(local_key_id2)
        if not usr1 or not usr2:
            # lg.warn('%r %r : not found sender or recipient key_id for %r' % (usr1, usr2, conv, ))
            return None
        usr1 = usr1.replace('master$', '')
        usr2 = usr2.replace('master$', '')
        idurl1 = global_id.glob2idurl(usr1, as_field=True)
        idurl2 = global_id.glob2idurl(usr2, as_field=True)
        conv_key_id = None
        conv_label = None
        user_idurl = None
        if (id_url.is_cached(idurl1) and idurl1 == my_id.getIDURL()) or usr1.split('@')[0] == my_id.getIDName():
            user_idurl = idurl2
            conv_key_id = global_id.UrlToGlobalID(idurl2, include_key=True)
            conv_label = conv_key_id.replace('master$', '').split('@')[0]
        if (id_url.is_cached(idurl2) and idurl2 == my_id.getIDURL()) or usr2.split('@')[0] == my_id.getIDName():
            user_idurl = idurl1
            conv_key_id = global_id.UrlToGlobalID(idurl1, include_key=True)
            conv_label = conv_key_id.replace('master$', '').split('@')[0]
        if conv_key_id:
            conv['key_id'] = conv_key_id
        if conv_label:
            conv['label'] = conv_label
        else:
            conv['label'] = conv_key_id
        if user_idurl:
            on_st = online_status.getInstance(user_idurl, autocreate=False)
            if on_st:
                conv.update(on_st.to_json())
    elif conv['type'] == 'group_message' or conv['type'] == 'personal_message':
        local_key_id, _, _ = conv['conversation_id'].partition('&')
        try:
            local_key_id = int(local_key_id)
        except:
            lg.exc()
            return None
        key_id = my_keys.get_local_key(local_key_id)
        if not key_id:
            # lg.warn('key_id was not found for %r' % conv)
            return None
        conv['key_id'] = key_id
        conv['label'] = my_keys.get_label(key_id) or key_id
        g_part = group_participant.get_active_group_participant(key_id)
        if g_part:
            conv.update(g_part.to_json())
    return conv


#------------------------------------------------------------------------------


def insert_message(data, message_id, message_time=None, sender=None, recipient=None, message_type=None, direction=None):
    """
    Writes JSON message to the message database.
    """
    payload_time = message_time or utime.utcnow_to_sec1970()
    payload_message_id = strng.to_text(message_id)
    payload_type = MESSAGE_TYPES.get(message_type, 1)
    if not sender:
        sender = my_id.getGlobalID(key_alias='master')
    if not recipient:
        recipient = my_id.getGlobalID(key_alias='master')
    if direction is None:
        if message_type in ['private_message', None]:
            direction = 'out' if sender == my_id.getGlobalID(key_alias='master') else 'in'
        else:
            direction = 'in'
    else:
        direction = direction.replace('incoming', 'in').replace('outgoing', 'out')
    if _Debug:
        lg.args(_DebugLevel, sender=sender, recipient=recipient, typ=payload_type, dir=direction, message_id=payload_message_id)
    recipient_local_key_id = my_keys.get_local_key_id(recipient)
    if payload_type in [3, 4]:
        sender_local_key_id = recipient_local_key_id
    else:
        sender_local_key_id = my_keys.get_local_key_id(sender)
    if sender_local_key_id is None or recipient_local_key_id is None:
        lg.err('failed to store message because local_key_id is not found, sender=%r recipient=%r' % (sender_local_key_id, recipient_local_key_id))
        return None
    cur().execute(
        '''INSERT INTO history (
            sender_local_key_id,
            sender_id,
            recipient_local_key_id,
            recipient_id,
            direction,
            payload_type,
            payload_time,
            payload_message_id,
            payload_body
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)''', (
            sender_local_key_id,
            sender,
            recipient_local_key_id,
            recipient,
            0 if direction == 'in' else 1,
            payload_type,
            payload_time,
            payload_message_id,
            data,
        )
    )
    db().commit()
    conversation_id = update_conversation(sender_local_key_id, recipient_local_key_id, payload_type, payload_time, payload_message_id)
    snap_id = '{}/{}'.format(conversation_id, payload_message_id)
    message_json = build_json_message(
        sender=sender,
        recipient=recipient,
        direction=direction,
        conversation_id=conversation_id,
        message_type=MESSAGE_TYPE_CODES.get(int(payload_type), 'private_message'),
        message_time=payload_time,
        message_id=payload_message_id,
        data=data,
    )
    listeners.push_snapshot('message', snap_id=snap_id, created=payload_time, data=message_json)
    return message_json


def update_conversation(sender_local_key_id, recipient_local_key_id, payload_type, payload_time, payload_message_id):
    conversation_id = get_conversation_id(sender_local_key_id, recipient_local_key_id, payload_type)
    if conversation_id is None:
        lg.err('failed to update conversation, local_key_id was not found')
        return None
    sql = 'SELECT * FROM conversations WHERE conversation_id=?'
    params = [
        conversation_id,
    ]
    found_conversation = list(cur().execute(sql, params))
    if found_conversation:
        sql = 'UPDATE conversations SET last_updated_time=?, last_message_id=? WHERE conversation_id=?'
        params = [
            payload_time,
            payload_message_id,
            conversation_id,
        ]
    else:
        sql = 'INSERT INTO conversations (conversation_id, payload_type, started_time, last_updated_time, last_message_id) VALUES (?, ?, ?, ?, ?)'
        params = [
            conversation_id,
            payload_type,
            payload_time,
            payload_time,
            payload_message_id,
        ]
    if _Debug:
        lg.args(_DebugLevel, conversation_id=conversation_id, found_conversation=len(found_conversation), params=params)
    cur().execute(sql, params)
    db().commit()
    if not found_conversation:
        snapshot = build_json_conversation(
            conversation_id=conversation_id,
            type=MESSAGE_TYPE_CODES.get(int(payload_type), 'private_message'),
            started=payload_time,
            last_updated=payload_time,
            last_message_id=payload_message_id,
        )
        listeners.push_snapshot('conversation', snap_id=conversation_id, data=snapshot)
    return conversation_id


def query_messages(sender_id=None, recipient_id=None, bidirectional=True, order_by_id=True, order_by_time=False, message_types=[], sequence_head=None, sequence_tail=None, offset=None, limit=None, raw_results=False):
    sql = 'SELECT * FROM history'
    q = ''
    params = []
    if bidirectional and sender_id and recipient_id:
        recipient_local_key_id = my_keys.get_local_key_id(recipient_id)
        sender_local_key_id = my_keys.get_local_key_id(sender_id)
        if recipient_local_key_id is None or sender_local_key_id is None:
            lg.warn('local_key_id was not found, recipient_local_key_id=%r sender_local_key_id=%r' % (recipient_local_key_id, sender_local_key_id))
            return []
        q += ' sender_local_key_id IN (?, ?) AND recipient_local_key_id IN (?, ?)'
        params += [
            sender_local_key_id,
            recipient_local_key_id,
            sender_local_key_id,
            recipient_local_key_id,
        ]
    else:
        if sender_id:
            sender_local_key_id = my_keys.get_local_key_id(sender_id)
            if sender_local_key_id is None:
                lg.warn('local_key_id was not found for sender %r' % sender_id)
                return []
            q += ' sender_local_key_id=?'
            params.append(sender_local_key_id)
        if recipient_id:
            recipient_local_key_id = my_keys.get_local_key_id(recipient_id)
            if recipient_local_key_id is None:
                lg.warn('local_key_id was not found for recipient %r' % recipient_id)
                return []
            q += ' recipient_local_key_id=?'
            params.append(recipient_local_key_id)
    if message_types:
        if params:
            q += ' AND payload_type IN (%s)' % (','.join([
                '?',
            ]*len(message_types)))
        else:
            q += ' payload_type IN (%s)' % (','.join([
                '?',
            ]*len(message_types)))
        params.extend([MESSAGE_TYPES.get(mt, 1) for mt in message_types])
    if sequence_head is not None:
        if params:
            q += ' AND payload_message_id>=?'
        else:
            q += ' payload_message_id>=?'
        params.append(sequence_head)
    if sequence_tail is not None:
        if params:
            q += ' AND payload_message_id<=?'
        else:
            q += ' payload_message_id<=?'
        params.append(sequence_tail)
    if q:
        sql += ' WHERE %s' % q
    if order_by_id:
        sql += ' ORDER BY payload_message_id DESC'
    else:
        if order_by_time:
            sql += ' ORDER BY payload_time DESC'
    if limit is not None:
        sql += ' LIMIT ?'
        params.append(limit)
    if offset is not None:
        sql += ' OFFSET ?'
        params.append(offset)
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    results = []
    local_key_ids = {}
    for row in cur().execute(sql, params):
        sender_local_key_id = row[0]
        sender_id_recorded = row[1]
        recipient_local_key_id = row[2]
        recipient_id_recorded = row[3]
        if sender_local_key_id not in local_key_ids:
            local_key_ids[sender_local_key_id] = my_keys.get_local_key(sender_local_key_id)
        if recipient_local_key_id not in local_key_ids:
            local_key_ids[recipient_local_key_id] = my_keys.get_local_key(recipient_local_key_id)
        if not local_key_ids.get(sender_local_key_id) or not local_key_ids.get(recipient_local_key_id):
            # lg.warn('unknown sender or recipient local key_id')
            continue
        if raw_results:
            if order_by_time:
                results.insert(0, row)
            else:
                results.append(row)
            continue
        json_msg = build_json_message(
            sender=sender_id_recorded,
            recipient=recipient_id_recorded,
            direction='in' if row[4] == 0 else 'out',
            conversation_id=get_conversation_id(sender_local_key_id, recipient_local_key_id, int(row[5])),
            message_type=MESSAGE_TYPE_CODES.get(int(row[5]), 'private_message'),
            message_time=row[6],
            message_id=row[7],
            data=json.loads(row[8]),
        )
        if order_by_time:
            results.insert(0, json_msg)
        else:
            results.append(json_msg)
    return results


def list_conversations(order_by_time=True, message_types=[], offset=None, limit=None):
    sql = 'SELECT * FROM conversations'
    q = ''
    params = []
    if message_types:
        q += ' payload_type IN (%s)' % (','.join([
            '?',
        ]*len(message_types)))
        params.extend([MESSAGE_TYPES.get(mt, 1) for mt in message_types])
    if q:
        sql += ' WHERE %s' % q
    if order_by_time:
        sql += ' ORDER BY last_updated_time DESC'
    if limit is not None:
        sql += ' LIMIT ?'
        params += [
            limit,
        ]
    if offset is not None:
        sql += ' OFFSET ?'
        params += [
            offset,
        ]
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    results = []
    for row in cur().execute(sql, params):
        results.append(dict(
            conversation_id=row[0],
            type=MESSAGE_TYPE_CODES.get(int(row[1]), 'private_message'),
            started=row[2],
            last_updated=row[3],
            last_message_id=row[4],
        ))
    return results


#------------------------------------------------------------------------------


def update_history_with_new_local_key_id(old_id, new_id):
    sql = 'UPDATE history SET sender_local_key_id=? WHERE sender_local_key_id=?'
    params = [
        new_id,
        old_id,
    ]
    cur().execute(sql, params)
    db().commit()
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    sql = 'UPDATE history SET recipient_local_key_id=? WHERE recipient_local_key_id=?'
    params = [
        new_id,
        old_id,
    ]
    cur().execute(sql, params)
    db().commit()
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)


def update_conversations_with_new_local_key_id(old_id, new_id):
    sql = 'SELECT * FROM conversations'
    params = []
    modifications = {}
    for row in cur().execute(sql, params):
        conversation_id = row[0]
        new_conversation_id = conversation_id
        id1, _, id2 = conversation_id.partition('&')
        try:
            id1 = int(id1)
            id2 = int(id2)
        except:
            lg.exc()
            continue
        if id1 == old_id:
            new_conversation_id = '{}&{}'.format(new_id, id2)
        if id2 == old_id:
            new_conversation_id = '{}&{}'.format(id1, new_id)
        if conversation_id != new_conversation_id:
            modifications[conversation_id] = new_conversation_id
    if _Debug:
        lg.args(_DebugLevel, modifications=modifications)
    for old_conv_id, new_conv_id in modifications.items():
        sql = 'UPDATE conversations SET conversation_id=? WHERE conversation_id=?'
        params = [
            new_conv_id,
            old_conv_id,
        ]
        cur().execute(sql, params)
        db().commit()
        if _Debug:
            lg.args(_DebugLevel, sql=sql, params=params)


def rebuild_conversations():
    cur().execute('SELECT count(name) FROM sqlite_master WHERE type="table" AND name="conversations"')
    if cur().fetchone()[0]:
        return
    cur().execute('''CREATE TABLE IF NOT EXISTS "conversations" (
        "conversation_id" TEXT,
        "payload_type" INTEGER,
        "started_time" INTEGER,
        "last_updated_time" INTEGER,
        "last_message_id" TEXT)''')
    cur().execute('CREATE INDEX "conversation id" on conversations(conversation_id)')
    db().commit()
    for message_json in list(query_messages()):
        msg_typ = message_json['payload'].get('msg_type') or message_json['payload'].get('type')
        payload_type = MESSAGE_TYPES.get(msg_typ, 1)
        recipient_local_key_id = my_keys.get_local_key_id(message_json['recipient']['glob_id'])
        if payload_type in [3, 4]:
            sender_local_key_id = recipient_local_key_id
        else:
            sender_local_key_id = my_keys.get_local_key_id(message_json['sender']['glob_id'])
        update_conversation(
            sender_local_key_id=sender_local_key_id,
            recipient_local_key_id=recipient_local_key_id,
            payload_type=payload_type,
            payload_time=message_json['payload']['time'],
            payload_message_id=message_json['payload']['message_id'],
        )


#------------------------------------------------------------------------------


def check_create_rename_key(new_public_key, new_key_id, new_local_key_id):
    conversation_type = None
    if new_key_id.startswith('group_'):
        conversation_type = 'group_message'
    elif new_key_id.startswith('person_'):
        conversation_type = 'personal_message'
    elif new_key_id.startswith('master$'):
        conversation_type = 'private_message'
    if conversation_type is None:
        return False
    sql = 'SELECT * FROM keys WHERE public_key=?'
    params = [
        new_public_key,
    ]
    found_public_keys = list(cur().execute(sql, params))
    if found_public_keys:
        if len(found_public_keys) > 1:
            raise Exception('found multiple records for same public key: %r' % found_public_keys)
        key_id = found_public_keys[0][0]
        local_key_id = found_public_keys[0][1]
        changed = False
        if key_id != new_key_id:
            changed = True
            sql = 'UPDATE keys SET key_id=? WHERE public_key=?'
            params = [
                new_key_id,
                new_public_key,
            ]
            cur().execute(sql, params)
            db().commit()
            if _Debug:
                lg.args(_DebugLevel, sql=sql, params=params)
        if local_key_id != new_local_key_id:
            changed = True
            lg.warn('found new public key which is re-using already known key with different local key id: %r -> %r' % (local_key_id, new_local_key_id))
            update_history_with_new_local_key_id(local_key_id, new_local_key_id)
            update_conversations_with_new_local_key_id(local_key_id, new_local_key_id)
            sql = 'UPDATE keys SET local_key_id=? WHERE public_key=?'
            params = [
                new_local_key_id,
                new_public_key,
            ]
            cur().execute(sql, params)
            db().commit()
            if _Debug:
                lg.args(_DebugLevel, sql=sql, params=params)
        return changed
    # new key will be registered in the "keys" table
    # also new conversation record can be created if that was a group type
    # unlike groups, private messages suppose to record a new conversation only when a first message appears and delivered between two
    sql = 'INSERT INTO keys (key_id, local_key_id, public_key) VALUES (?, ?, ?)'
    params = [
        new_key_id,
        new_local_key_id,
        new_public_key,
    ]
    cur().execute(sql, params)
    db().commit()
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    if conversation_type in ['group_message', 'personal_message']:
        update_conversation(
            sender_local_key_id=new_local_key_id,
            recipient_local_key_id=new_local_key_id,
            payload_type=MESSAGE_TYPES.get(conversation_type, 1),
            payload_time=utime.utcnow_to_sec1970(),
            payload_message_id='',
        )
    return True


#------------------------------------------------------------------------------


def fetch_conversations(order_by_time=True, message_types=[], offset=None, limit=None):
    conversations = []
    for conv_record in list(list_conversations(
        order_by_time=order_by_time,
        message_types=message_types,
        offset=offset,
        limit=limit,
    ), ):
        conv = build_json_conversation(**conv_record)
        if conv:
            if conv['key_id']:
                conversations.append(conv)
            else:
                lg.warn('unknown key_id for %r' % conv)
    return conversations


def populate_conversations(message_types=[], offset=0, limit=100, order_by_time=True):
    for conv in fetch_conversations(
        order_by_time=order_by_time,
        message_types=message_types,
        offset=offset,
        limit=limit,
    ):
        listeners.push_snapshot('conversation', snap_id=conv['conversation_id'], data=conv)


def populate_messages(recipient_id=None, sender_id=None, message_types=[], offset=0, limit=100):
    if recipient_id:
        if not recipient_id.count('@'):
            from bitdust.contacts import contactsdb
            recipient_idurl = contactsdb.find_correspondent_by_nickname(recipient_id)
            if not recipient_idurl:
                lg.err('recipient %r was not found' % recipient_id)
                return
            recipient_id = global_id.UrlToGlobalID(recipient_idurl)
        recipient_glob_id = global_id.ParseGlobalID(recipient_id)
        if not recipient_glob_id['idurl']:
            lg.err('wrong recipient_id')
            return
        recipient_id = global_id.MakeGlobalID(**recipient_glob_id)
        if not my_keys.is_valid_key_id(recipient_id):
            lg.err('invalid recipient_id: %s' % recipient_id)
            return
    if sender_id:
        sender_local_key_id = my_keys.get_local_key_id(sender_id)
        if sender_local_key_id is None:
            return
    if recipient_id:
        recipient_local_key_id = my_keys.get_local_key_id(recipient_id)
        if recipient_local_key_id is None:
            lg.warn('recipient %r local key id was not registered' % recipient_id)
            return
    for row in query_messages(
        sender_id=sender_id,
        recipient_id=recipient_id,
        bidirectional=False,
        message_types=message_types,
        offset=offset,
        limit=limit,
        raw_results=True,
    ):
        conversation_id = get_conversation_id(row[0], row[2], int(row[5]))
        if conversation_id is None:
            continue
        snap_id = '{}/{}'.format(conversation_id, row[7])
        snapshot = build_json_message(
            sender=row[1],
            recipient=row[3],
            direction='in' if row[4] == 0 else 'out',
            conversation_id=conversation_id,
            message_type=MESSAGE_TYPE_CODES.get(int(row[5]), 'private_message'),
            message_time=row[6],
            message_id=row[7],
            data=json.loads(row[8]),
        )
        listeners.push_snapshot('message', snap_id=snap_id, created=row[6], data=snapshot)


#------------------------------------------------------------------------------


def main():
    import pprint
    my_keys.init()
    init()
    pprint.pprint(list(query_messages(
        # sender_id='',
        recipient_id='group_96b1fa942b556e2e0dbccdcc9eeed84a$veselibro@seed.bitdust.io',
        # offset=2,
        # limit=3,
        message_types=[
            'group_message',
        ],
    )))
    # pprint.pprint(list(list_conversations(
    # message_types=['group_message', ]
    # )))
    shutdown()


if __name__ == '__main__':
    lg.set_debug_level(20)
    main()



#!/usr/bin/env python
# udp_connector.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (udp_connector.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: udp_connector.

.. role:: red
BitDust udp_connector() Automat


EVENTS:
    * :red:`abort`
    * :red:`dht-read-failed`
    * :red:`dht-read-success`
    * :red:`dht-write-failed`
    * :red:`dht-write-success`
    * :red:`start`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

import time

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.dht import dht_service

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

_ConnectorsDict = {}

#------------------------------------------------------------------------------


def connectors():
    """

    """
    global _ConnectorsDict
    return _ConnectorsDict


def create(node, peer_id):
    if _Debug:
        lg.out(_DebugLevel, 'udp_connector.create peer_id=%s' % peer_id)
    c = DHTUDPConnector(node, peer_id)
    connectors()[c.id] = c
    return c


def get(peer_id):
    for c in connectors().values():
        if c.peer_id == peer_id:
            return c
    return None


#------------------------------------------------------------------------------


class DHTUDPConnector(automat.Automat):
    """
    This class implements all the functionality of the ``udp_connector()``
    state machine.
    """

    fast = True

    def __init__(self, node, peer_id):
        self.node = node
        self.peer_id = peer_id
        name = 'udp_connector[%s]' % self.peer_id
        automat.Automat.__init__(self, name, 'AT_STARTUP', 18)

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.listen_port = None
        self.my_id = None
        self.my_address = None
        self.working_deferred = None

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'start':
                self.state = 'DHT_LOOP'
                self.doInit(*args, **kwargs)
                self.KeyPosition = 0
                self.doDHTReadIncoming(*args, **kwargs)
        #---DHT_WRITE---
        elif self.state == 'DHT_WRITE':
            if event == 'dht-write-success':
                self.state = 'DHT_READ'
                self.doDHTReadPeerAddress(*args, **kwargs)
            elif event == 'dht-write-failed':
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'abort':
                self.state = 'ABORTED'
                self.doDestroyMe(*args, **kwargs)
        #---DHT_READ---
        elif self.state == 'DHT_READ':
            if event == 'dht-read-success':
                self.state = 'DONE'
                self.doStartNewSession(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-read-failed':
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'abort':
                self.state = 'ABORTED'
                self.doDestroyMe(*args, **kwargs)
        #---DONE---
        elif self.state == 'DONE':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        #---DHT_LOOP---
        elif self.state == 'DHT_LOOP':
            if event == 'dht-read-failed':
                self.state = 'DHT_WRITE'
                self.doDHTWriteIncoming(*args, **kwargs)
            elif event == 'dht-read-success' and self.KeyPosition >= 10:
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-read-success' and self.KeyPosition < 10 and not self.isMyIncoming(*args, **kwargs):
                self.KeyPosition += 1
                self.doDHTReadIncoming(*args, **kwargs)
            elif event == 'dht-read-success' and self.isMyIncoming(*args, **kwargs):
                self.state = 'DHT_READ'
                self.doDHTReadPeerAddress(*args, **kwargs)
            elif event == 'abort':
                self.state = 'ABORTED'
                self.doDestroyMe(*args, **kwargs)
        #---ABORTED---
        elif self.state == 'ABORTED':
            pass
        return None

    def isMyIncoming(self, *args, **kwargs):
        """
        Condition method.
        """
        incoming_peer_id, incoming_user_address = args[0]
        return incoming_peer_id == self.my_id and incoming_user_address == self.my_address

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.listen_port, self.my_id, self.my_address = args[0]

    def doDHTReadIncoming(self, *args, **kwargs):
        """
        Action method.
        """
        key = self.peer_id + ':incoming' + str(self.KeyPosition)
        self.working_deferred = dht_service.get_value(key)
        if not self.working_deferred:
            self.automat('dht-read-failed')
        else:
            self.working_deferred.addCallback(self._got_peer_incoming, key, self.KeyPosition)
            self.working_deferred.addErrback(lambda x: self.automat('dht-read-failed'))

    def doDHTWriteIncoming(self, *args, **kwargs):
        """
        Action method.
        """
        key = self.peer_id + ':incoming' + str(self.KeyPosition)
        value = '%s %s:%d %s\n' % (str(self.my_id), self.my_address[0], self.my_address[1], str(time.time()))
        if _Debug:
            lg.out(_DebugLevel, 'doDHTWriteIncoming  key=%s' % key)
        self.working_deferred = dht_service.set_value(key, value, age=int(time.time()))
        if not self.working_deferred:
            self.automat('dht-write-failed')
        else:
            try:
                self.working_deferred.addCallback(self._wrote_peer_incoming)
                self.working_deferred.addErrback(lambda x: self.automat('dht-write-failed'))
            except:
                self.automat('dht-write-failed')

    def doStartNewSession(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.udp import udp_session
        peer_address = args[0]
        if self.node.my_address is None:
            if _Debug:
                lg.out(_DebugLevel, 'udp_connector.doStartNewSession to %s at %s SKIP because my_address is None' % (self.peer_id, peer_address))
            return
        active_sessions = udp_session.get(peer_address)
        if active_sessions:
            if _Debug:
                lg.out(_DebugLevel, 'udp_connector.doStartNewSession SKIP because found existing by peer address %s : %s' % (peer_address, active_sessions))
            return
        active_sessions = udp_session.get_by_peer_id(self.peer_id)
        if active_sessions:
            if _Debug:
                lg.out(_DebugLevel, 'udp_connector.doStartNewSession SKIP because found existing by peer id %s : %s' % (self.peer_id, active_sessions))
            return
        s = udp_session.create(self.node, peer_address, self.peer_id)
        s.automat('init', (self.listen_port, self.my_id, self.my_address))

    def doDHTReadPeerAddress(self, *args, **kwargs):
        """
        Action method.
        """
        key = self.peer_id + ':address'
        self.working_deferred = dht_service.get_value(key)
        if not self.working_deferred:
            self.automat('dht-read-failed')
        else:
            self.working_deferred.addCallback(self._got_peer_address, key)
            self.working_deferred.addErrback(lambda x: self.automat('dht-read-failed'))

    def doReportFailed(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.udp import udp_session
        udp_session.report_and_remove_pending_outbox_files_to_host(self.peer_id, 'unable to establish connection')

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        if self.working_deferred:
            self.working_deferred.cancel()
            self.working_deferred = None
        self.node = None
        connectors().pop(self.id)
        self.destroy()

    def _got_peer_incoming(self, value, key, position):
        if _Debug:
            lg.out(_DebugLevel, 'udp_connector._got_peer_incoming at position %d: %d' % (position, len(str(value))))
        self.working_deferred = None
        incoming = None
        if not isinstance(value, dict):
            self.automat('dht-read-failed')
            return
        try:
            # incoming = value.values()[0]
            incoming = value[dht_service.key_to_hash(key)]
        except:
            lg.out(2, '%r' % value)
            lg.exc()
            self.automat('dht-read-failed')
            return
        try:
            incoming_peer_id, incoming_user_address, _ = incoming.split(b' ')
            incoming_user_address = incoming_user_address.split(b':')
            incoming_user_address = (incoming_user_address[0], int(incoming_user_address[1]))
        except:
            lg.out(2, '%r' % incoming)
            lg.exc()
            self.automat('dht-read-failed')
            return
        self.automat('dht-read-success', (incoming_peer_id, incoming_user_address))

    def _wrote_peer_incoming(self, nodes):
        self.working_deferred = None
        if len(nodes) > 0:
            self.automat('dht-write-success')
        else:
            self.automat('dht-write-failed')

    def _got_peer_address(self, value, key):
        if not isinstance(value, dict):
            self.automat('dht-read-failed')
            return
        try:
            peer_ip, peer_port = value[dht_service.key_to_hash(key)].split(b':')
            peer_port = int(peer_port)
        except:
            lg.exc()
            self.automat('dht-read-failed')
            return
        if _Debug:
            lg.out(_DebugLevel, 'udp_connector._got_peer_address %s:%d ~ %s' % (peer_ip, peer_port, self.peer_id))
        self.automat('dht-read-success', (peer_ip, peer_port))




class Fork():
    def __init__(self):
        self.POW_FORK = 1450000
        self.POW_FORK_TESTNET = 894170
        self.FORK_AHEAD = 5
        self.versions_remove = ['mainnet0020', 'mainnet0019', 'mainnet0018', 'mainnet0017']
        self.FORK_REWARD = None
        self.FORK_REWARD_TESTNET = None
        self.PASSED = False
        self.PASSED_TESTNET = False
        self.REWARD_MAX = 6
        #self.POW_FORK = 1168860 #HACK
        #self.versions_remove = [] #HACK
        #self.REWARD_MAX = 5 #HACK

    def limit_version(self, node):
        for allowed_version in node.version_allow:
            if allowed_version in self.versions_remove:
                node.logger.app_log.warning(f'Beginning to reject old protocol versions - block {node.last_block}')
                node.version_allow.remove(allowed_version)

    def check_postfork_reward(self, db_handler):
        # ram
        try:
            db_handler.execute_param(db_handler.c, 'SELECT reward FROM transactions WHERE block_height = ? AND reward != 0', (self.POW_FORK + 1, ))
            self.FORK_REWARD = db_handler.c.fetchone()[0]

        except:
            # hdd in case we have not saved yet
            db_handler.execute_param(db_handler.h, 'SELECT reward FROM transactions WHERE block_height = ? AND reward != 0', (self.POW_FORK + 1, ))
            self.FORK_REWARD = db_handler.h.fetchone()[0]

        if self.FORK_REWARD < self.REWARD_MAX:
            self.PASSED = True
        return self.PASSED

    def check_postfork_reward_testnet(self, db_handler):
        #ram
        try:
            db_handler.execute_param(db_handler.c, 'SELECT reward FROM transactions WHERE block_height = ? AND reward != 0', (self.POW_FORK_TESTNET + 1, ))
            self.FORK_REWARD_TESTNET = db_handler.c.fetchone()[0]
        except:
            #hdd in case we have not saved yet
            db_handler.execute_param(db_handler.h, 'SELECT reward FROM transactions WHERE block_height = ? AND reward != 0', (self.POW_FORK_TESTNET + 1, ))
            self.FORK_REWARD_TESTNET = db_handler.h.fetchone()[0]

        # print(type(self.FORK_REWARD_TESTNET))

        if self.FORK_REWARD_TESTNET < self.REWARD_MAX:
            self.PASSED_TESTNET = True
        return self.PASSED_TESTNET



#!/usr/bin/python
# fire_hire.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (fire_hire.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: fire_hire

.. role:: red

.. raw:: html

    <a href="https://bitdust.io/automats/fire_hire/fire_hire.png" target="_blank">
    <img src="https://bitdust.io/automats/fire_hire/fire_hire.png" style="max-width:100%;">
    </a>

If at some point we are not getting good answers from a node
for too long we need to replace him and reconstruct the data
he was holding. This is fire & hire and then rebuilding and scrubbing.

Probably if we try to contact someone for 48 hours and can not,
we want to give up on them.

User can use GUI to fire & hire at any time.

Automatically fire if right after we ask a supplier for a BigPacket,
he turns around and asks us for it (like he does not have it).

Our regular code would not do this, but an evil modified version might
try to get away with not holding any data by just getting it from us
any time we asked for it. So we can not allow this cheat to work.

We ask for lists of files they have for us and keep these in
``settings.SuppliersDir()/[supplier idurl]``
These should be updated at least every night.
If a supplier has not given us a list for several days
he is a candidate for firing.

Gateway code keeps statistics on how fast different nodes are.
We could fire a slowest node periodically.

Restore can keep track of who did not answer in time to be part of raidread,
and they can be a candidate for firing.

The fire packet needs to use IDURL so that if there is a retransmission of the "fire" request
we just send new "list suppliers" again.

Task list
1) fire inactive suppliers (default is 48 hours)
2) fire suppliers with low rating (less than 25% by default)
3) test if supplier is "evil modified"
4) test ListFiles periodically
5) fire slow nodes

The ``fire_hire()`` automat is our "human resources department".
The objective is to discharge the outsiders among suppliers -
those who are offline for too long or who holds the least of our data.

Each time the machine is started,
it chooses the weakest supplier and sends a request to the Central server
to the dismissal of that supplier.
In response, the Central server sends a new list
in which there is a new supplier instead dismissed.

If all providers are reliable enough - machine will not work and will reset.


EVENTS:
    * :red:`init`
    * :red:`instant`
    * :red:`made-decision`
    * :red:`restart`
    * :red:`search-failed`
    * :red:`supplier-connected`
    * :red:`supplier-state-changed`
    * :red:`timer-25sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys
import time

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in fire_hire.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import global_state
from bitdust.automats import automat

from bitdust.lib import misc
from bitdust.lib import diskspace
from bitdust.lib import strng
from bitdust.lib import utime

from bitdust.main import config
from bitdust.main import settings
from bitdust.main import events

from bitdust.contacts import contactsdb

from bitdust.services import driver

from bitdust.raid import eccmap

from bitdust.userid import my_id
from bitdust.userid import id_url

#-------------------------------------------------------------------------

_FireHire = None
_LastFireTime = 0
_SuppliersToFire = []
_ForceRestartState = False

#-------------------------------------------------------------------------


def GetLastFireTime():
    """
    This method returns a time moment when last time some supplier was
    replaced.
    """
    global _LastFireTime
    return _LastFireTime


def ClearLastFireTime():
    _LastFireTime = 0


def AddSupplierToFire(idurl):
    global _SuppliersToFire
    _SuppliersToFire.append(idurl)


def ForceRestart():
    global _ForceRestartState
    _ForceRestartState = True


def IsAllHired():
    if settings.getSuppliersNumberDesired() < 0:
        # I must know how many suppliers I want
        lg.warn('my desired number of suppliers not set')
        return False
    if contactsdb.num_suppliers() != settings.getSuppliersNumberDesired():
        # I must have exactly that amount of suppliers already
        if _Debug:
            lg.args(_DebugLevel, desiried_suppliers=settings.getSuppliersNumberDesired(), current_suppliers=contactsdb.num_suppliers())
        return False
    if id_url.is_some_empty(contactsdb.suppliers()):
        # I must know all of my suppliers
        if _Debug:
            lg.args(_DebugLevel, my_suppliers=contactsdb.suppliers())
        return False
    return True


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _FireHire
    if _FireHire is None:
        _FireHire = FireHire(
            name='fire_hire',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _FireHire.automat(event, *args, **kwargs)
    return _FireHire


def Destroy():
    """
    Destroy the state machine and remove the instance from memory.
    """
    global _FireHire
    if _FireHire is None:
        return
    _FireHire.destroy()
    del _FireHire
    _FireHire = None


#------------------------------------------------------------------------------


class FireHire(automat.Automat):

    """
    This class implements all the functionality of the ``fire_hire()`` state
    machine.
    """

    fast = False

    timers = {
        'timer-25sec': (25.0, ['FIRE_MANY', 'SUPPLIERS?']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.connect_list = []
        self.dismiss_list = []
        self.dismiss_results = []
        self.hire_list = []
        self.configs = (None, None)
        self.restart_interval = 1.0
        self.restart_task = None

    def shutdown(self):
        self.connect_list = []
        self.dismiss_list = []
        self.dismiss_results = []
        self.hire_list = []
        self.configs = (None, None)
        self.restart_interval = 1.0
        self.restart_task = None

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when automat's state was changed.
        """
        global_state.set_global_state('FIREHIRE ' + newstate)
        if newstate == 'READY' and event != 'instant':
            self.automat('instant')

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'READY'
                self.NeedRestart = False
        #---READY---
        elif self.state == 'READY':
            if (event == 'restart' or (event == 'instant' and self.NeedRestart)) and (self.isConfigChanged(*args, **kwargs) or self.isContractExpiring(*args, **kwargs)) and self.isExistSomeSuppliers(*args, **kwargs):
                self.state = 'SUPPLIERS?'
                self.NeedRestart = False
                self.doSaveConfig(*args, **kwargs)
                self.doConnectSuppliers(*args, **kwargs)
            elif (event == 'restart' or (event == 'instant' and self.NeedRestart)) and not ((self.isConfigChanged(*args, **kwargs) or self.isContractExpiring(*args, **kwargs)) and self.isExistSomeSuppliers(*args, **kwargs)):
                self.state = 'DECISION?'
                self.NeedRestart = False
                self.doDecideToDismiss(*args, **kwargs)
        #---DECISION?---
        elif self.state == 'DECISION?':
            if event == 'made-decision' and self.isSomeoneToDismiss(*args, **kwargs) and not self.isMoreNeeded(*args, **kwargs):
                self.state = 'FIRE_MANY'
                self.doRememberSuppliers(*args, **kwargs)
                self.doRemoveSuppliers(*args, **kwargs)
                self.doDisconnectSuppliers(*args, **kwargs)
            elif event == 'restart':
                self.NeedRestart = True
            elif event == 'made-decision' and not self.isMoreNeeded(*args, **kwargs) and not self.isSomeoneToDismiss(*args, **kwargs):
                self.state = 'READY'
                self.doNotifyFinished(*args, **kwargs)
            elif event == 'made-decision' and self.isMoreNeeded(*args, **kwargs):
                self.state = 'HIRE_ONE'
                self.doRememberSuppliers(*args, **kwargs)
                self.doFindNewSupplier(*args, **kwargs)
        #---HIRE_ONE---
        elif self.state == 'HIRE_ONE':
            if event == 'restart':
                self.NeedRestart = True
            elif event == 'supplier-connected' and not self.isStillNeeded(*args, **kwargs) and self.isSomeoneToDismiss(*args, **kwargs):
                self.state = 'FIRE_MANY'
                self.doSubstituteSupplier(*args, **kwargs)
                self.doDisconnectSuppliers(*args, **kwargs)
            elif event == 'supplier-connected' and not self.isStillNeeded(*args, **kwargs) and not self.isSomeoneToDismiss(*args, **kwargs):
                self.state = 'READY'
                self.doSubstituteSupplier(*args, **kwargs)
                self.doNotifySuppliersChanged(*args, **kwargs)
            elif event == 'supplier-connected' and self.isStillNeeded(*args, **kwargs):
                self.doSubstituteSupplier(*args, **kwargs)
                self.doFindNewSupplier(*args, **kwargs)
            elif event == 'search-failed' and not self.isSomeoneToDismiss(*args, **kwargs):
                self.state = 'READY'
                self.doScheduleNextRestart(*args, **kwargs)
                self.doNotifySuppliersChanged(*args, **kwargs)
            elif event == 'search-failed' and self.isSomeoneToDismiss(*args, **kwargs):
                self.state = 'FIRE_MANY'
                self.doDisconnectSuppliers(*args, **kwargs)
                self.doRemoveSuppliers(*args, **kwargs)
                self.doScheduleNextRestart(*args, **kwargs)
        #---FIRE_MANY---
        elif self.state == 'FIRE_MANY':
            if event == 'supplier-state-changed' and not self.isAllDismissed(*args, **kwargs):
                self.doCloseConnector(*args, **kwargs)
            elif event == 'restart':
                self.NeedRestart = True
            elif event == 'supplier-state-changed' and self.isAllDismissed(*args, **kwargs):
                self.state = 'READY'
                self.doCloseConnector(*args, **kwargs)
                self.doClearDismissList(*args, **kwargs)
                self.doNotifySuppliersChanged(*args, **kwargs)
            elif event == 'timer-25sec':
                self.state = 'READY'
                self.doCloseConnectors(*args, **kwargs)
                self.doClearDismissList(*args, **kwargs)
                self.doNotifySuppliersChanged(*args, **kwargs)
        #---SUPPLIERS?---
        elif self.state == 'SUPPLIERS?':
            if event == 'restart':
                self.NeedRestart = True
            elif (event == 'supplier-state-changed' and self.isAllReady(*args, **kwargs)) or event == 'timer-25sec':
                self.state = 'DECISION?'
                self.doDecideToDismiss(*args, **kwargs)
        return None

    def isMoreNeeded(self, *args, **kwargs):
        """
        Condition method.
        """
        # if _Debug:
        #     lg.out(_DebugLevel, 'fire_hire.isMoreNeeded current=%d dismiss=%d needed=%d' % (
        #         contactsdb.num_suppliers(), len(self.dismiss_list),
        #         settings.getSuppliersNumberDesired()))
        if id_url.is_some_empty(contactsdb.suppliers()):
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.isMoreNeeded found empty supplier!!!')
            return True
        if isinstance(args[0], list):
            dismissed = args[0]
        else:
            dismissed = self.dismiss_list
        s = set(id_url.to_bin_list(contactsdb.suppliers()))
        s.difference_update(set(id_url.to_bin_list(dismissed)))
        result = len(s) < settings.getSuppliersNumberDesired()
        if _Debug:
            lg.out(_DebugLevel, 'fire_hire.isMoreNeeded %d %d %d %d, result=%s' % (contactsdb.num_suppliers(), len(dismissed), len(s), settings.getSuppliersNumberDesired(), result))
        return result

    def isAllReady(self, *args, **kwargs):
        """
        Condition method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'fire_hire.isAllReady %d %d' % (len(self.connect_list), contactsdb.num_suppliers()))
        return len(self.connect_list) == 0  # contactsdb.num_suppliers()

    def isAllDismissed(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.dismiss_list) == len(self.dismiss_results)

    def isSomeoneToDismiss(self, *args, **kwargs):
        """
        Condition method.
        """
        if args and isinstance(args[0], list):
            dismissed = args[0]
        else:
            dismissed = self.dismiss_list
        return len(dismissed) > 0

    def isStillNeeded(self, *args, **kwargs):
        """
        Condition method.
        """
        supplier_idurl = args[0]
        current_suppliers = contactsdb.suppliers()
        if supplier_idurl in current_suppliers:
            # this guy is already a supplier, we still need more then
            return True
        desired_number = settings.getSuppliersNumberDesired()
        needed_suppliers = current_suppliers[:desired_number]
        empty_suppliers = needed_suppliers.count(id_url.field(b''))
        # if '' in needed_suppliers:
        # lg.warn('found empty suppliers!!!')
        # return True
        s = set(id_url.to_bin_list(needed_suppliers))
        s.add(id_url.to_bin(supplier_idurl))
        s.difference_update(set(id_url.to_bin_list(self.dismiss_list)))
        result = len(s) - empty_suppliers < settings.getSuppliersNumberDesired()
        # if _Debug:
        #     lg.out(_DebugLevel, 'fire_hire.isStillNeeded %d %d %d %d %d, result=%s' % (
        #     contactsdb.num_suppliers(), len(needed_suppliers), len(self.dismiss_list),
        #     len(s), settings.getSuppliersNumberDesired(), result))
        return result

    def isConfigChanged(self, *args, **kwargs):
        """
        Condition method.
        """
        global _ForceRestartState
        if _ForceRestartState:
            return True
        return self._is_config_changed()

    def isContractExpiring(self, *args, **kwargs):
        """
        Condition method.
        """
        if not driver.is_on('service_customer_contracts'):
            return False
        return len(self._get_suppliers_with_expired_contracts()) > 0

    def isExistSomeSuppliers(self, *args, **kwargs):
        """
        Condition method.
        """
        sup_list = contactsdb.suppliers()
        return contactsdb.num_suppliers() > 0 and sup_list.count(id_url.field(b'')) < contactsdb.num_suppliers()

    def doSaveConfig(self, *args, **kwargs):
        """
        Action method.
        """
        self.configs = (
            settings.getSuppliersNumberDesired(),
            diskspace.GetBytesFromString(settings.getNeededString()),
        )

    def doConnectSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        global _ForceRestartState
        _ForceRestartState = False
        from bitdust.customer import supplier_connector
        from bitdust.p2p import online_status
        self.connect_list = []
        my_current_family = contactsdb.suppliers()
        target_suppliers = list(my_current_family)
        if not self._is_config_changed():
            if driver.is_on('service_customer_contracts'):
                # only select suppliers with expired contracts
                target_suppliers = self._get_suppliers_with_expired_contracts()
        if _Debug:
            lg.args(_DebugLevel, my_current_family=my_current_family, target_suppliers=target_suppliers, configs=self.configs)
        for pos, supplier_idurl in enumerate(my_current_family):
            if not supplier_idurl:
                continue
            if self.configs[0] and pos >= self.configs[0]:
                continue
            if not id_url.is_in(supplier_idurl, target_suppliers):
                continue
            sc = supplier_connector.by_idurl(supplier_idurl)
            if sc is None:
                sc = supplier_connector.create(
                    supplier_idurl=supplier_idurl,
                    customer_idurl=my_id.getIDURL(),
                )
            else:
                sc.needed_bytes = None
                sc.do_calculate_needed_bytes()
            sc.set_callback('fire_hire', self._on_supplier_connector_state_changed)
            self.connect_list.append(supplier_idurl)
            sc.automat(
                'connect',
                family_position=pos,
                ecc_map=eccmap.Current().name,
                family_snapshot=id_url.to_bin_list(my_current_family),
            )
            online_status.add_online_status_listener_callback(
                idurl=supplier_idurl,
                callback_method=self._on_supplier_online_status_state_changed,
            )

    def doDecideToDismiss(self, *args, **kwargs):
        """
        Action method.
        """
        global _SuppliersToFire
        from bitdust.p2p import p2p_connector
        from bitdust.p2p import network_connector
        from bitdust.customer import supplier_connector
        from bitdust.p2p import online_status
        # take any actions only if I am connected to the network
        if not p2p_connector.A() or not network_connector.A():
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.doDecideToDismiss    p2p_connector() is not ready yet, SKIP')
            self.automat('made-decision', [])
            return
        if not network_connector.A():
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.doDecideToDismiss    network_connector() is not ready yet, SKIP')
            self.automat('made-decision', [])
            return
        if p2p_connector.A().state != 'CONNECTED' or network_connector.A().state != 'CONNECTED':
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.doDecideToDismiss    p2p/network is not connected at the moment, SKIP')
            self.automat('made-decision', [])
            return
        # if certain suppliers needs to be removed by manual/external request just do that
        to_be_fired = id_url.to_list(set(_SuppliersToFire))
        _SuppliersToFire = []
        if to_be_fired:
            lg.info('going to fire %d suppliers from external request' % len(to_be_fired))
            self.automat('made-decision', to_be_fired)
            return
        # make sure to not go too far when i just want to decrease number of my suppliers
        number_desired = settings.getSuppliersNumberDesired()
        redundant_suppliers = set()
        if contactsdb.num_suppliers() > number_desired:
            for supplier_index in range(number_desired, contactsdb.num_suppliers()):
                idurl = contactsdb.supplier(supplier_index)
                if idurl:
                    lg.info('found REDUNDANT supplier %s at position %d' % (idurl, supplier_index))
                    redundant_suppliers.add(idurl)
        if redundant_suppliers:
            result = list(redundant_suppliers)
            lg.info('will replace redundant suppliers: %s' % result)
            self.automat('made-decision', result)
            return
        # now I need to look more careful at my suppliers
        potentialy_fired = set()
        connected_suppliers = set()
        disconnected_suppliers = set()
        requested_suppliers = set()
        online_suppliers = set()
        offline_suppliers = set()
        # if you have some empty suppliers need to get rid of them,
        # but no need to dismiss anyone at the moment.
        my_suppliers = contactsdb.suppliers()
        if _Debug:
            lg.args(_DebugLevel, my_suppliers=my_suppliers)
        if id_url.is_some_empty(my_suppliers):
            lg.warn('SKIP, found empty supplier')
            self.automat('made-decision', [])
            return
        for supplier_idurl in my_suppliers:
            sc = supplier_connector.by_idurl(supplier_idurl)
            if not sc:
                lg.warn('SKIP, supplier connector for supplier %s not exist' % supplier_idurl)
                continue
            if sc.state == 'NO_SERVICE':
                lg.warn('found "NO_SERVICE" supplier: %s' % supplier_idurl)
                disconnected_suppliers.add(supplier_idurl)
                potentialy_fired.add(supplier_idurl)
            elif sc.state == 'CONNECTED':
                connected_suppliers.add(supplier_idurl)
            elif sc.state in ['DISCONNECTED', 'REFUSE']:
                disconnected_suppliers.add(supplier_idurl)
#             elif sc.state in ['QUEUE?', 'REQUEST', ]:
#                 requested_suppliers.add(supplier_idurl)
            if online_status.isOffline(supplier_idurl):
                offline_suppliers.add(supplier_idurl)
            elif online_status.isOnline(supplier_idurl):
                online_suppliers.add(supplier_idurl)
            elif online_status.isCheckingNow(supplier_idurl):
                requested_suppliers.add(supplier_idurl)
        if not connected_suppliers or not online_suppliers:
            lg.warn('SKIP, no ONLINE suppliers found at the moment')
            self.automat('made-decision', [])
            return
        if requested_suppliers:
            lg.warn('SKIP, still waiting response from some of suppliers')
            self.automat('made-decision', [])
            return
        if not disconnected_suppliers:
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.doDecideToDismiss    SKIP, no OFFLINE suppliers found at the moment')
            # TODO: add more conditions to fire "slow" suppliers - they are still connected but useless
            self.automat('made-decision', [])
            return
        if len(offline_suppliers) + len(online_suppliers) != number_desired:
            lg.warn('SKIP, offline + online != total count: %s %s %s' % (offline_suppliers, online_suppliers, number_desired))
            self.automat('made-decision', [])
            return
        max_offline_suppliers_count = eccmap.GetCorrectableErrors(number_desired)
        if len(offline_suppliers) > max_offline_suppliers_count:
            lg.warn('SKIP, too many OFFLINE suppliers at the moment : %d > %d' % (len(offline_suppliers), max_offline_suppliers_count))
            self.automat('made-decision', [])
            return
        critical_offline_suppliers_count = eccmap.GetFireHireErrors(number_desired)
        if len(offline_suppliers) >= critical_offline_suppliers_count and len(offline_suppliers) > 0:
            if config.conf().getBool('services/employer/replace-critically-offline-enabled'):
                # TODO: check that issue
                # too aggressive replacing suppliers who still have the data is very dangerous !!!
                one_dead_supplier = offline_suppliers.pop()
                lg.warn('found "CRITICALLY_OFFLINE" supplier %s, max offline limit is %d' % (one_dead_supplier, critical_offline_suppliers_count))
                potentialy_fired.add(one_dead_supplier)
        if not potentialy_fired:
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.doDecideToDismiss   found no "bad" suppliers, all is good !!!!!')
            self.automat('made-decision', [])
            return
        # only replace suppliers one by one at the moment
        result = list(potentialy_fired)
        lg.info('will replace supplier %s' % result[0])
        self.automat('made-decision', [
            result[0],
        ])

    def doRememberSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        self.dismiss_list = id_url.to_list(args[0])

    def doFindNewSupplier(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'fire_hire.doFindNewSupplier desired_suppliers=%d current_suppliers=%r' % (settings.getSuppliersNumberDesired(), contactsdb.suppliers()))
        from bitdust.p2p import network_connector
        if network_connector.A().state != 'CONNECTED':
            if _Debug:
                lg.out(_DebugLevel, '        network_connector is not CONNECTED at the moment, SKIP')
            self.automat('search-failed')
            return
        position_for_new_supplier = None
        for pos in range(settings.getSuppliersNumberDesired()):
            if pos in self.hire_list:
                continue
            supplier_idurl = contactsdb.supplier(pos)
            if not supplier_idurl:
                lg.info('found empty supplier at position %d and going to find new supplier on that position' % pos)
                position_for_new_supplier = pos
                break
            if id_url.is_in(supplier_idurl, self.dismiss_list, as_field=False):
                lg.info('going to find new supplier on existing position %d to replace supplier %s' % (pos, supplier_idurl))
                position_for_new_supplier = pos
                break
        if position_for_new_supplier is None:
            lg.err('did not found position for new supplier')
            self.automat('search-failed')
            return
        from bitdust.customer import supplier_finder
        for idurl_txt in strng.to_text(config.conf().getString('services/employer/candidates')).split(','):
            if idurl_txt.strip():
                supplier_finder.AddSupplierToHire(idurl_txt)
        self.hire_list.append(position_for_new_supplier)
        supplier_finder.A(
            'start',
            family_position=position_for_new_supplier,
            ecc_map=eccmap.Current().name,
            family_snapshot=id_url.to_bin_list(contactsdb.suppliers()),
        )

    def doSubstituteSupplier(self, *args, **kwargs):
        """
        Action method.
        """
        new_idurl = id_url.field(args[0])
        family_position = kwargs.get('family_position')
        current_suppliers = list(contactsdb.suppliers())
        desired_suppliers = settings.getSuppliersNumberDesired()
        old_idurl = None
        if family_position in self.hire_list:
            self.hire_list.remove(family_position)
            lg.info('found position on which new supplier suppose to be hired: %d' % family_position)
        else:
            lg.warn('did not found position for new supplier to be hired on')
        if new_idurl in current_suppliers:
            raise Exception('%s is already supplier' % new_idurl)
        if family_position is None or family_position == -1:
            lg.warn('unknown family_position from supplier results, will pick first empty spot')
            position = -1
            old_idurl = None
            for i in range(len(current_suppliers)):
                if not current_suppliers[i].strip():
                    position = i
                    break
                if id_url.is_in(current_suppliers[i], self.dismiss_list, as_field=False):
                    position = i
                    old_idurl = current_suppliers[i]
                    break
            family_position = position
        if _Debug:
            lg.out(_DebugLevel, 'fire_hire.doSubstituteSupplier family_position=%d' % family_position)
        contactsdb.add_supplier(idurl=new_idurl, position=family_position)
        contactsdb.save_suppliers()
        misc.writeSupplierData(
            new_idurl,
            'connected',
            time.strftime('%d-%m-%Y %H:%M:%S'),
            my_id.getIDURL(),
        )
        if family_position < 0:
            lg.info('added new supplier, family position unknown: %s desired_suppliers=%d current_suppliers=%d' % (new_idurl, desired_suppliers, len(contactsdb.suppliers())))
            # yapf: disable
            events.send('supplier-modified', data=dict(
                new_idurl=new_idurl,
                old_idurl=None,
                position=family_position,
                ecc_map=eccmap.Current().name,
                family_snapshot=id_url.to_bin_list(contactsdb.suppliers()),
            ))
            # yapf: enable
        else:
            if old_idurl:
                lg.info('hired new supplier and substitute existing supplier on position %d : %s->%s desired_suppliers=%d current_suppliers=%d' % (family_position, old_idurl, new_idurl, desired_suppliers, len(contactsdb.suppliers())))
                # yapf: disable
                events.send('supplier-modified', data=dict(
                    new_idurl=new_idurl,
                    old_idurl=old_idurl,
                    position=family_position,
                    ecc_map=eccmap.Current().name,
                    family_snapshot=id_url.to_bin_list(contactsdb.suppliers()),
                ))
                # yapf: enable
            else:
                lg.info('hired new supplier on empty position %d : %s desired_suppliers=%d current_suppliers=%d' % (family_position, new_idurl, desired_suppliers, len(contactsdb.suppliers())))
                # yapf: disable
                events.send('supplier-modified', data=dict(
                    new_idurl=new_idurl,
                    old_idurl=None,
                    position=family_position,
                    ecc_map=eccmap.Current().name,
                    family_snapshot=id_url.to_bin_list(contactsdb.suppliers()),
                ))
                # yapf: enable
        self.restart_interval = 1.0
        if _Debug:
            lg.out(_DebugLevel, '    my current suppliers: %r' % contactsdb.suppliers())

    def doRemoveSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        current_suppliers = contactsdb.suppliers()
        desired_suppliers = settings.getSuppliersNumberDesired()
        if len(current_suppliers) < desired_suppliers:
            lg.warn('must have more suppliers %d<%d' % (len(current_suppliers), desired_suppliers))
        removed_suppliers = []
        for supplier_idurl in self.dismiss_list:
            if id_url.is_not_in(supplier_idurl, current_suppliers, as_field=False):
                lg.warn('%s not a supplier' % supplier_idurl)
                continue
            pos = current_suppliers.index(id_url.field(supplier_idurl))
            current_suppliers[pos] = ''
            removed_suppliers.append((
                pos,
                supplier_idurl,
            ))
            misc.writeSupplierData(
                supplier_idurl,
                'disconnected',
                time.strftime('%d-%m-%Y %H:%M:%S'),
                my_id.getIDURL(),
            )
        current_suppliers = current_suppliers[:desired_suppliers]
        contactsdb.update_suppliers(current_suppliers)
        contactsdb.save_suppliers()
        for position, supplier_idurl in removed_suppliers:
            events.send('supplier-modified', data=dict(
                new_idurl=None,
                old_idurl=supplier_idurl,
                position=position,
            ))
        lg.info('removed some suppliers : %d  desired_suppliers=%d current_suppliers=%d' % (len(self.dismiss_list), desired_suppliers, len(contactsdb.suppliers())))
        if _Debug:
            lg.out(_DebugLevel, '    my current suppliers: %r' % contactsdb.suppliers())

    def doDisconnectSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import supplier_connector
        from bitdust.p2p import online_status
        if _Debug:
            lg.out(_DebugLevel, 'fire_hire.doDisconnectSuppliers %r' % self.dismiss_list)
        self.dismiss_results = []
        for supplier_idurl in self.dismiss_list:
            sc = supplier_connector.by_idurl(supplier_idurl)
            if sc:
                sc.set_callback('fire_hire', self._on_supplier_connector_state_changed)
                sc.automat('disconnect', ecc_map=eccmap.Current().name)
            else:
                lg.warn('supplier_connector must exist, but not found %s' % supplier_idurl)
                self.dismiss_results.append(supplier_idurl)
            online_status.remove_online_status_listener_callback(
                idurl=supplier_idurl,
                callback_method=self._on_supplier_online_status_state_changed,
            )

    def doCloseConnector(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import supplier_connector
        supplier_idurl, _ = args[0]
        supplier_idurl = id_url.field(supplier_idurl)
        if _Debug:
            lg.args(_DebugLevel, supplier_idurl=supplier_idurl, dismiss_list=self.dismiss_list)
        sc = supplier_connector.by_idurl(supplier_idurl)
        if id_url.is_in(supplier_idurl.to_bin(), self.dismiss_list, as_field=False):
            self.dismiss_list.remove(supplier_idurl)
        if sc:
            sc.automat('shutdown')
        else:
            raise Exception('supplier_connector must exist')

    def doCloseConnectors(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import supplier_connector
        for supplier_idurl in self.dismiss_list:
            sc = supplier_connector.by_idurl(supplier_idurl)
            if id_url.is_in(supplier_idurl, self.dismiss_list, as_field=False):
                self.dismiss_list.remove(supplier_idurl)
            if sc:
                sc.automat('shutdown')

    def doClearDismissList(self, *args, **kwargs):
        """
        Action method.
        """
        self.dismiss_list = []

    def doScheduleNextRestart(self, *args, **kwargs):
        """
        Action method.
        """
        self.hire_list = []
        if not self.restart_task:
            self.restart_task = reactor.callLater(  # @UndefinedVariable
                self.restart_interval,
                self._scheduled_restart,
            )
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.doScheduleNextRestart after %r sec.' % self.restart_interval)
            from bitdust.p2p import network_connector
            if network_connector.A().state != 'CONNECTED':
                self.restart_interval = 60*5
            else:
                self.restart_interval *= 1.1
        else:
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire.doScheduleNextRestart already scheduled - %r sec. left' % (time.time() - self.restart_task.getTime()))

    def doNotifySuppliersChanged(self, *args, **kwargs):
        self.hire_list = []
        if driver.is_on('service_backups'):
            from bitdust.storage import backup_monitor
            backup_monitor.A('suppliers-changed')

    def doNotifyFinished(self, *args, **kwargs):
        self.hire_list = []
        if driver.is_on('service_backups'):
            from bitdust.storage import backup_monitor
            backup_monitor.A('fire-hire-finished')

    def _is_config_changed(self):
        if None in self.configs:
            return True
        curconfigs = (settings.getSuppliersNumberDesired(), diskspace.GetBytesFromString(settings.getNeededString()))
        return self.configs[0] != curconfigs[0] or self.configs[1] != curconfigs[1]

    def _get_suppliers_with_expired_contracts(self):
        from bitdust.customer import supplier_connector
        now = utime.utcnow_to_sec1970()
        ret = []
        for supplier_idurl in contactsdb.suppliers():
            if supplier_idurl:
                sc = supplier_connector.by_idurl(supplier_idurl)
                if sc:
                    if sc.state == 'CONNECTED':
                        if sc.storage_contract:
                            if now > utime.unpack_time(sc.storage_contract['complete_after']):
                                lg.warn('storage contract with %s is expired' % supplier_idurl)
                                ret.append(supplier_idurl)
        return ret

    def _scheduled_restart(self):
        self.restart_task = None
        self.automat('restart')

    def _on_supplier_connector_state_changed(self, idurl, newstate, **kwargs):
        from bitdust.customer import supplier_connector
        idurl = id_url.field(idurl)
        if _Debug:
            lg.out(_DebugLevel, 'fire_hire._on_supplier_connector_state_changed %s to %s, own state is %s ' % (idurl, newstate, self.state))
        if supplier_connector.by_idurl(idurl):
            supplier_connector.by_idurl(idurl).remove_callback('fire_hire', self._on_supplier_connector_state_changed)
        if self.state == 'SUPPLIERS?':
            if idurl in self.connect_list:
                self.connect_list.remove(idurl)
            else:
                lg.warn('did not found %r in connect_list' % idurl)
        elif self.state == 'FIRE_MANY':
            if idurl not in self.dismiss_results:
                self.dismiss_results.append(idurl)
            else:
                lg.warn('did not found %r in dismiss_results' % idurl)
        else:
            return
        self.automat('supplier-state-changed', (
            idurl,
            newstate,
        ))

    def _on_supplier_online_status_state_changed(self, oldstate, newstate, event_string, *args, **kwargs):
        if oldstate != newstate:
            if _Debug:
                lg.out(_DebugLevel, 'fire_hire._on_supplier_online_status_state_changed  %s -> %s, own state is %s' % (oldstate, newstate, self.state))
        if oldstate != newstate and newstate in [
            'CONNECTED',
            'OFFLINE',
        ]:
            self.automat('restart')


# def WhoIsLost():
#    """
#    If we have more than 50% data packets lost to someone and it was a long story - fire this guy.
#    We check this first, because this is more important than other things.
#    Well, many things can be a reason:
#    * refuse storage service,
#    * disconnected,
#    * slow connection,
#    * old code,
#    * too many network errors,
#    * timeout during sending
#    If we can not send him our data or retrieve it back - how can we do a backups to him even if he is online?
#    """
#
#    for supplier_idurl in contactsdb.suppliers():
#    # for sc in supplier_connector.connectors().values():
#        sc = supplier_connector.by_idurl(supplier_idurl)
#        if sc and sc.state == 'NO_SERVICE':
#            lg.out(6, 'fire_hire.WhoIsLost !!!!!!!! %s : no service' % sc.idurl)
#            return 'found-one-lost-supplier', sc.idurl
#    unreliable_supplier = None
#    most_fails = 0.0
#    for supplierNum in range(contactsdb.num_suppliers()):
#        idurl = contactsdb.supplier(supplierNum)
#        if not idurl:
#            continue
#        if not data_sender.statistic().has_key(idurl):
#            continue
#        stats = data_sender.statistic()[idurl]
#        total = stats[0] + stats[1]
#        failed = stats[1]
#        if total > 10:
#            failed_percent = failed / total
#            if failed_percent > 0.5:
#                if most_fails < failed_percent:
#                    most_fails = failed_percent
#                    unreliable_supplier = idurl
#    if unreliable_supplier:
#        return 'found-one-lost-supplier', unreliable_supplier

#    # we only fire offline suppliers
#    offline_suppliers = {}

#    # ask backup_monitor about current situation
#    # check every offline supplier and see how many files he keep at the moment
#    for supplierNum in range(contactsdb.num_suppliers()):
#        idurl = contactsdb.supplier(supplierNum)
#        if not idurl:
#            continue
#        if contact_status.isOnline(idurl):
#            continue
#        blocks, total, stats = backup_matrix.GetSupplierStats(supplierNum)
#        rating = 0 if total == 0 else blocks / total
#        offline_suppliers[idurl] = rating

#    # if all suppliers are online - we are very happy - no need to fire anybody!
#    if len(offline_suppliers) == 0:
#        lg.out(4, 'fire_hire.WhoIsLost no offline suppliers, Cool!')
#        return 'not-found-lost-suppliers', ''

#    # sort users - we always fire worst supplier
#    rating = offline_suppliers.keys()
#    rating.sort(key=lambda idurl: offline_suppliers[idurl])
#    lost_supplier_idurl = rating[0]

#    # we do not want to fire this man if he store at least 50% of our files
#    # the fact that he is offline is not enough to fire him!
#    if offline_suppliers[lost_supplier_idurl] < 0.5 and backup_fs.sizebackups() > 0:
#        lg.out(4, 'fire_hire.WhoIsLost !!!!!!!! %s is offline and keeps only %d%% of our data' % (
#            nameurl.GetName(lost_supplier_idurl),
#            int(offline_suppliers[lost_supplier_idurl] * 100.0)))
#        return 'found-one-lost-supplier', lost_supplier_idurl

#    # but if we did not saw him for a long time - we do not want him for sure
#    if time.time() - ratings.connected_time(lost_supplier_idurl) > 60 * 60 * 24 * 2:
#        lg.out(2, 'fire_hire.WhoIsLost !!!!!!!! %s is offline and keeps %d%% of our data, but he was online %d hours ago' % (
#            nameurl.GetName(lost_supplier_idurl),
#            int(offline_suppliers[lost_supplier_idurl] * 100.0),
#            int((time.time() - ratings.connected_time(lost_supplier_idurl)) * 60 * 60),))
#        return 'found-one-lost-supplier', lost_supplier_idurl

#    lg.out(2, 'fire_hire.WhoIsLost some people is not here, but we did not found the bad guy at this time')
#    return 'not-found-lost-suppliers', ''



#!/usr/bin/python
# service_customer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_customer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_customer
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return CustomerService()


class CustomerService(LocalService):

    service_name = 'service_customer'
    config_path = 'services/customer/enabled'

    def dependent_on(self):
        return [
            'service_data_disintegration',
        ]

    def installed(self):
        from bitdust.userid import my_id
        if not my_id.isLocalIdentityReady():
            return False
        return True

    def start(self):
        from bitdust.contacts import contactsdb
        from bitdust.customer import supplier_connector
        from bitdust.userid import my_id
        from bitdust.main import events
        for _, supplier_idurl in enumerate(contactsdb.suppliers()):
            if supplier_idurl and not supplier_connector.by_idurl(supplier_idurl, customer_idurl=my_id.getIDURL()):
                supplier_connector.create(
                    supplier_idurl=supplier_idurl,
                    customer_idurl=my_id.getIDURL(),
                )
        events.add_subscriber(self._on_my_keys_synchronized, 'my-keys-synchronized')
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        # TODO: read from dht and connect to other suppliers - from other customers who shared data to me
        return True

    def stop(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.customer import supplier_connector
        from bitdust.userid import my_id
        from bitdust.main import events
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        events.remove_subscriber(self._on_my_keys_synchronized, 'my-keys-synchronized')
        for sc in supplier_connector.connectors(my_id.getIDURL()).values():
            reactor.callLater(0, sc.automat, 'shutdown')  # @UndefinedVariable
        # TODO: disconnect other suppliers
        return True

    def health_check(self):
        from bitdust.customer import supplier_connector
        from bitdust.userid import my_id
        for sc in supplier_connector.connectors(my_id.getIDURL()).values():
            # at least one supplier must be online to consider my customer service to be healthy
            if sc.state in [
                'CONNECTED',
            ]:
                return True
        return False

    def _on_my_keys_synchronized(self, evt):
        from bitdust.logs import lg
        from bitdust.userid import my_id
        # customer_key_id = my_id.getGlobalID(key_alias='customer')
        # if not my_keys.is_key_registered(customer_key_id) and keys_synchronizer.is_synchronized():
        #     lg.info('customer key was not found but we know that all keys are in sync, generate new key: %s' % customer_key_id)
        #     my_keys.generate_key(customer_key_id, key_size=settings.getPrivateKeySize())

    def _on_identity_url_changed(self, evt):
        from bitdust.logs import lg
        from bitdust.userid import id_url
        from bitdust.contacts import contactsdb
        from bitdust.customer import supplier_connector
        old_idurl = id_url.field(evt.data['old_idurl'])
        for customer_idurl, suppliers_list in contactsdb.all_suppliers(as_dict=True).items():
            if old_idurl == customer_idurl:
                customer_idurl.refresh()
                lg.info('found customer family idurl rotated : %r -> %r' % (evt.data['old_idurl'], evt.data['new_idurl']))
            for supplier_pos, supplier_idurl in enumerate(suppliers_list):
                if old_idurl == supplier_idurl:
                    supplier_idurl.refresh()
                    lg.info('found supplier idurl rotated for customer family %r at position %r : %r -> %r' % (customer_idurl, supplier_pos, evt.data['old_idurl'], evt.data['new_idurl']), )
            for customer_idurl, sc_dict in supplier_connector.connectors(as_dict=True).items():
                if old_idurl == customer_idurl:
                    customer_idurl.refresh()
                    lg.info('found customer idurl rotated in supplier_connector.connectors() : %r -> %r' % (evt.data['old_idurl'], evt.data['new_idurl']))
                for supplier_idurl, sc in sc_dict.items():
                    if old_idurl == supplier_idurl:
                        supplier_idurl.refresh()
                        sc.customer_idurl.refresh()
                        sc.supplier_idurl.refresh()
                        lg.info('found supplier idurl rotated in %r for customer %r : %r -> %r' % (sc, customer_idurl, evt.data['old_idurl'], evt.data['new_idurl']))



import log
import re


def replace_regex(string, replace):
    replaced_string = re.sub(r'^{}'.format(replace), '', string)
    return replaced_string


def aliases_update(node, db_handler_instance):
    db_handler_instance.index_cursor.execute('SELECT block_height FROM aliases ORDER BY block_height DESC LIMIT 1;')
    try:
        alias_last_block = int(db_handler_instance.index_cursor.fetchone()[0])
    except:
        alias_last_block = 0

    node.logger.app_log.warning('Alias anchor block: {}'.format(alias_last_block))

    db_handler_instance.h.execute('SELECT block_height, address, openfield FROM transactions WHERE openfield LIKE ? AND block_height >= ? ORDER BY block_height ASC, timestamp ASC;', ('alias=' + '%', ) + (alias_last_block, ))
    # include the anchor block in case indexation stopped there
    result = db_handler_instance.h.fetchall()

    for openfield in result:
        alias = (replace_regex(openfield[2], 'alias='))
        node.logger.app_log.warning(f'Processing alias registration: {alias}')
        try:
            db_handler_instance.index_cursor.execute('SELECT * from aliases WHERE alias = ?', (alias, ))
            dummy = db_handler_instance.index_cursor.fetchall()[0]  # check for uniqueness
            node.logger.app_log.warning(f'Alias already registered: {alias}')
        except:
            db_handler_instance.index_cursor.execute('INSERT INTO aliases VALUES (?,?,?)', (openfield[0], openfield[1], alias))
            db_handler_instance.index.commit()
            node.logger.app_log.warning(f'Added alias to the database: {alias} from block {openfield[0]}')


if __name__ == '__main__':
    app_log = log.log('aliases.log', 'WARNING', True)



#!/usr/bin/env python
# stun_client.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (stun_client.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: stun_client.

.. role:: red

BitDust ``stun_client()`` Automat


EVENTS:
    * :red:`datagram-received`
    * :red:`dht-nodes-not-found`
    * :red:`found-some-nodes`
    * :red:`init`
    * :red:`port-number-received`
    * :red:`shutdown`
    * :red:`start`
    * :red:`timer-10sec`
    * :red:`timer-1sec`
    * :red:`timer-2sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys
import random
import re

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.automats import automat

from bitdust.lib import strng
from bitdust.lib import udp
from bitdust.lib import net_misc
from bitdust.lib import nameurl
from bitdust.lib import misc

from bitdust.main import settings
from bitdust.main import events

from bitdust.services import driver

#------------------------------------------------------------------------------

_StunClient = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _StunClient
    if event is None:
        return _StunClient
    if _StunClient is None:
        # set automat name and starting state here
        _StunClient = StunClient(
            name='stun_client',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _StunClient.automat(event, *args, **kwargs)
    return _StunClient


class StunClient(automat.Automat):
    """
    This class implements all the functionality of the ``stun_client()`` state
    machine.
    """
    # fast = True

    timers = {
        'timer-1sec': (1.0, ['REQUEST']),
        'timer-2sec': (2.0, ['REQUEST']),
        'timer-10sec': (10.0, ['PORT_NUM?', 'REQUEST']),
    }

    MESSAGES = {
        'MSG_01': 'not found any DHT nodes',
        'MSG_02': 'not found any available stun servers',
        'MSG_03': 'timeout responding from stun servers',
    }

    def msg(self, msgid, *args, **kwargs):
        return self.MESSAGES.get(msgid, '')

    def init(self):
        self.listen_port = None
        self.callbacks = []
        self.find_nodes_attempts = 1
        self.minimum_needed_servers = 1
        self.minimum_needed_results = 1
        self.stun_nodes = []
        self.stun_servers = []
        self.stun_results = {}
        self.my_address = None
        self.deferreds = {}

    def getMyExternalAddress(self):
        return self.my_address

    def dropMyExternalAddress(self):
        self.my_address = None

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'STOPPED'
                self.doInit(*args, **kwargs)
        #---STOPPED---
        elif self.state == 'STOPPED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start':
                self.state = 'RANDOM_NODES'
                self.doAddCallback(*args, **kwargs)
                self.doDHTFindRandomNode(*args, **kwargs)
        #---REQUEST---
        elif self.state == 'REQUEST':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-2sec':
                self.doStun(*args, **kwargs)
            elif event == 'timer-10sec' and not self.isSomeServersResponded(*args, **kwargs):
                self.state = 'STOPPED'
                self.doReportFailed(self.msg('MSG_03', *args, **kwargs))
                self.doClearResults(*args, **kwargs)
            elif event == 'start':
                self.doAddCallback(*args, **kwargs)
            elif event == 'datagram-received' and self.isMyIPPort(*args, **kwargs) and self.isNeedMoreResults(*args, **kwargs):
                self.doRecordResult(*args, **kwargs)
            elif event == 'port-number-received':
                self.doAddStunServer(*args, **kwargs)
                self.doStun(*args, **kwargs)
            elif (event == 'timer-1sec' and self.isSomeServersResponded(*args, **kwargs)) or (event == 'datagram-received' and self.isMyIPPort(*args, **kwargs) and not self.isNeedMoreResults(*args, **kwargs)):
                self.state = 'KNOW_MY_IP'
                self.doRecordResult(*args, **kwargs)
                self.doReportSuccess(*args, **kwargs)
                self.doClearResults(*args, **kwargs)
        #---KNOW_MY_IP---
        elif self.state == 'KNOW_MY_IP':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start':
                self.state = 'RANDOM_NODES'
                self.doAddCallback(*args, **kwargs)
                self.doDHTFindRandomNode(*args, **kwargs)
        #---RANDOM_NODES---
        elif self.state == 'RANDOM_NODES':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-nodes-not-found':
                self.state = 'STOPPED'
                self.doReportFailed(self.msg('MSG_01', *args, **kwargs))
            elif event == 'start':
                self.doAddCallback(*args, **kwargs)
            elif event == 'found-some-nodes' and self.isNeedMoreNodes(*args, **kwargs):
                self.doRememberStunNodes(*args, **kwargs)
                self.doDHTFindRandomNode(*args, **kwargs)
            elif event == 'found-some-nodes' and not self.isNeedMoreNodes(*args, **kwargs):
                self.state = 'PORT_NUM?'
                self.doRememberStunNodes(*args, **kwargs)
                self.doRequestStunPortNumbers(*args, **kwargs)
        #---PORT_NUM?---
        elif self.state == 'PORT_NUM?':
            if event == 'start':
                self.doAddCallback(*args, **kwargs)
            elif event == 'timer-10sec':
                self.state = 'STOPPED'
                self.doReportFailed(self.msg('MSG_02', *args, **kwargs))
                self.doClearResults(*args, **kwargs)
            elif event == 'port-number-received':
                self.state = 'REQUEST'
                self.doAddStunServer(*args, **kwargs)
                self.doStun(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isMyIPPort(self, *args, **kwargs):
        """
        Condition method.
        """
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            return False
        return command == udp.CMD_MYIPPORT

    def isSomeServersResponded(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.stun_results) > 0

    def isNeedMoreResults(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.stun_results) <= self.minimum_needed_results

    def isNeedMoreNodes(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.stun_nodes) + len(*args, **kwargs) < self.minimum_needed_servers

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.listen_port = args[0]
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.doInit on port %d' % self.listen_port)
        if udp.proto(self.listen_port):
            udp.proto(self.listen_port).add_callback(self._datagram_received)
        else:
            lg.warn('udp port %s is not opened' % self.listen_port)

    def doAddCallback(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            self.callbacks.append(args[0])

    def doDHTFindRandomNode(self, *args, **kwargs):
        """
        Action method.
        """
        self._find_random_node()

    def doRememberStunNodes(self, *args, **kwargs):
        """
        Action method.
        """
        nodes = args[0]
        for node in nodes:
            if node not in self.stun_nodes:
                self.stun_nodes.append(node)

    def doRequestStunPortNumbers(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel + 4, 'stun_client.doRequestStunPortNumbers')
        for node in self.stun_nodes:
            if node.id in self.deferreds:
                lg.warn('Already requested stun_port from %r' % node)
                continue
            if _Debug:
                lg.out(_DebugLevel + 4, '    from %s' % node)
            d = node.request('stun_port')
            d.addBoth(self._stun_port_received, node)
            self.deferreds[node.id] = d

    def doAddStunServer(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel + 4, 'stun_client.doAddStunServer %s' % str(*args, **kwargs))
        self.stun_servers.append(args[0])

    def doStun(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0] is not None:
            if _Debug:
                lg.out(_DebugLevel + 4, 'stun_client.doStun to one stun_server: %s' % str(*args, **kwargs))
            udp.send_command(self.listen_port, udp.CMD_STUN, b'', *args, **kwargs)
            return
        if _Debug:
            lg.out(_DebugLevel + 4, 'stun_client.doStun to %d stun_servers' % (len(self.stun_servers)))  # , self.stun_servers))
        for address in self.stun_servers:
            if address is None:
                continue
            if address in list(self.stun_results.keys()):
                continue
            udp.send_command(self.listen_port, udp.CMD_STUN, b'', address)

    def doRecordResult(self, *args, **kwargs):
        """
        Action method.
        """
        if not args or args[0] is None:
            return
        try:
            datagram, address = args[0]
            command, payload = datagram
            ip, port = payload.split(b':')
            port = int(port)
        except:
            lg.exc()
        self.stun_results[address] = (ip, port)
        # if len(self.stun_results) >= len(self.stun_servers):
        #     self.automat('all-responded')

    def doClearResults(self, *args, **kwargs):
        """
        Action method.
        """
        self.stun_nodes = []
        self.stun_servers = []
        self.stun_results = {}

    def doReportSuccess(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            min_port = min([addr[1] for addr in list(self.stun_results.values())])
            max_port = max([addr[1] for addr in list(self.stun_results.values())])
            my_ip = strng.to_text(list(self.stun_results.values())[0][0])
            if min_port == max_port:
                result = ('stun-success', 'non-symmetric', my_ip, min_port)
            else:
                result = ('stun-success', 'symmetric', my_ip, self.stun_results)
            self.my_address = (my_ip, min_port)
        except:
            lg.exc()
            result = ('stun-failed', None, None, [])
            self.my_address = None
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.doReportSuccess based on %d nodes: %r' % (len(self.stun_results), result))
        if self.my_address:
            current_external_ip = misc.readExternalIP()
            if current_external_ip != self.my_address[0]:
                events.send('my-external-ip-changed', data=dict(
                    old=current_external_ip,
                    new=self.my_address[0],
                ))
            bpio.WriteTextFile(settings.ExternalIPFilename(), self.my_address[0])
            bpio.WriteTextFile(settings.ExternalUDPPortFilename(), str(self.my_address[1]))
        for cb in self.callbacks:
            cb(result[0], result[1], result[2], result[3])
        self.callbacks = []

    def doReportFailed(self, *args, **kwargs):
        """
        Action method.
        """
        self.my_address = None
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.doReportFailed : %s' % args[0])
        for cb in self.callbacks:
            cb('stun-failed', None, None, [])
        self.callbacks = []

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        global _StunClient
        _StunClient = None
        for d in list(self.deferreds.values()):
            if d and not d.called:
                d.cancel()
        self.deferreds.clear()
        if udp.proto(self.listen_port):
            udp.proto(self.listen_port).remove_callback(self._datagram_received)
        self.destroy()

    def _datagram_received(self, datagram, address):
        self.automat('datagram-received', (
            datagram,
            address,
        ))
        return False

    def _find_random_nodes(self, tries, result_list, prev_key=None):
        if prev_key and prev_key in self.deferreds:
            self.deferreds.pop(prev_key)
        if _Debug:
            lg.out(_DebugLevel + 4, 'stun_client._find_random_nodes tries=%d result_list=%d' % (tries, len(result_list)))
        if tries <= 0 or len(result_list) >= self.minimum_needed_servers:
            if len(result_list) > 0:
                self.automat('found-some-nodes', result_list)
            else:
                self.automat('dht-nodes-not-found')
            return
        from bitdust.dht import dht_service
        new_key = dht_service.random_key()
        d = dht_service.find_node(new_key)
        d.addCallback(lambda nodes: self._find_random_nodes(tries - 1, list(set(result_list + nodes)), new_key))
        d.addErrback(lambda x: self._find_random_nodes(tries - 1, result_list, new_key))
        self.deferreds[new_key] = d

    def _find_random_node(self):
        if _Debug:
            lg.out(_DebugLevel + 4, 'stun_client._find_random_node')
        from bitdust.dht import dht_service
        new_key = dht_service.random_key()
        d = dht_service.find_node(new_key)
        d.addCallback(self._some_nodes_found)
        d.addErrback(self._nodes_not_found)
        # self.deferreds[new_key] = d

    def _some_nodes_found(self, nodes):
        if _Debug:
            lg.out(_DebugLevel + 4, 'stun_client._some_nodes_found : %r' % nodes)
        if nodes and len(nodes) > 0:
            self.automat('found-some-nodes', nodes)
        else:
            self.automat('dht-nodes-not-found')

    def _nodes_not_found(self, err):
        if _Debug:
            lg.out(_DebugLevel, 'stun_client._nodes_not_found err=%s' % str(err))
        self.automat('dht-nodes-not-found')

    def _stun_port_received(self, result, node):
        if _Debug:
            lg.out(_DebugLevel, 'stun_client._stun_port_received  %r from %s node_id=%r' % (result, node, node.id))
        self.deferreds.pop(node.id, None)
        if not isinstance(result, dict):
            if _Debug:
                lg.dbg(_DebugLevel, 'empty result received from node %r : %r' % (node, result))
            return
        try:
            port = int(strng.to_text(result['stun_port']))
            address = node.address
        except:
            lg.exc()
            return
        if port == 0:
            return
        if _Debug:
            lg.out(_DebugLevel, '        new stun port server found  %s:%s' % (address, port))
        self.automat('port-number-received', (address, port))


#------------------------------------------------------------------------------


def udp_dht_stun(udp_port=None, dht_port=None, result_defer=None):
    if not driver.is_on('service_my_ip_port'):
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.udp_dht_stun   SKIP because service_my_ip_port() is not started')
        if result_defer:
            result_defer.errback(Exception('service_my_ip_port() is not started'))
        return False

    from bitdust.dht import dht_service

    if dht_service.node().connectingTask() and not dht_service.node().connectingTask().called:
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.udp_dht_stun   SKIP and run later because dht_service is still joining the network')
        dht_service.node().connectingTask().addCallback(lambda ok: udp_dht_stun(udp_port=udp_port, dht_port=dht_port, result_defer=result_defer))
        if result_defer:
            dht_service.node().connectingTask().addErrback(result_defer.errback)
        return True

    dht_port = dht_port or settings.getDHTPort()
    udp_port = udp_port or settings.getUDPPort()
    if udp_port:
        udp.listen(udp_port)
    if dht_port:
        dht_service.init(dht_port)
    d = dht_service.connect()

    def _cb(cod, typ, ip, details):
        # A('shutdown')
        ret = {
            'result': cod,  # 'stun-success' or 'stun-failed'
            'type': typ,
            'ip': ip,
            'details': details,
        }
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.udp_dht_stun   result : %r' % ret)
        result_defer.callback(ret)
        return None

    def _go(live_nodes):
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.udp_dht_stun   GO with nodes: %r' % live_nodes)
        A('init', udp_port)
        A('start', _cb)
        return live_nodes

    d.addCallback(_go)
    if result_defer:
        d.addErrback(result_defer.errback)
        # d.addErrback(lambda err: result_defer.callback(dict(ip='127.0.0.1', errors=[str(err), ])))
    return True


def http_stun(result_defer=None):
    from bitdust.userid import known_servers
    identity_servers = known_servers.by_host()
    if not identity_servers:
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.http_stun   SKIP, no known identity servers found')
        return False
    one_host = random.choice(list(identity_servers.keys()))
    one_port = identity_servers[one_host][0]
    one_url = nameurl.UrlMake('http', one_host, one_port)
    if _Debug:
        lg.out(_DebugLevel, 'stun_client.http_stun   GO with one node : %r' % one_url)

    def _check_body(html_response):
        ret = {
            'result': 'stun-failed',
            'type': None,
            'ip': None,
            'details': ['unknown client host from response'],
        }
        mo = re.search(b'\<\!\-\-CLIENT_HOST\=([\d\.]+):(\d+)\-\-\>', html_response)
        if not mo:
            if _Debug:
                lg.out(_DebugLevel, 'stun_client.http_stun   FAILED : unknown client host from response')
            if result_defer:
                result_defer.callback(ret)
            return None
        ret = {
            'result': 'stun-success',
            'type': 'unknown',
            'ip': strng.to_text(mo.group(1)),
            'details': [],
        }
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.http_stun   SUCCESS : %r' % ret)
        if result_defer:
            result_defer.callback(ret)
        return None

    def _request_failed(err):
        if _Debug:
            lg.out(_DebugLevel, 'stun_client.http_stun   FAILED : %r' % err)
        ret = {
            'result': 'stun-failed',
            'type': None,
            'ip': None,
            'details': [err.getErrorMessage()],
        }
        if result_defer:
            result_defer.callback(ret)
        return None

    d = net_misc.getPageTwisted(one_url)
    d.addCallback(_check_body)
    d.addErrback(_request_failed)
    return True


def safe_stun(udp_port=None, dht_port=None, result_defer=None):
    from twisted.internet.defer import Deferred
    result = result_defer or Deferred()

    def _check_response(response):
        if response.get('result') == 'stun-success':
            result.callback(response)
            return None
        http_stun(result_defer=result)
        return None

    def _fallback(err):
        http_stun(result_defer=result)
        return None

    first_result = Deferred()
    first_result.addCallback(_check_response)
    first_result.addErrback(_fallback)
    udp_dht_stun(udp_port, dht_port, result_defer=first_result)
    return result


def test_safe_stun():
    from twisted.internet import reactor  # @UnresolvedImport

    def _cb(res):
        print(res)
        reactor.stop()  # @UndefinedVariable

    def _eb(err):
        print(err)
        reactor.stop()  # @UndefinedVariable

    lg.set_debug_level(30)
    safe_stun().addCallbacks(_cb, _eb)
    reactor.run()  # @UndefinedVariable


#------------------------------------------------------------------------------


def main():
    from twisted.internet import reactor  # @UnresolvedImport
    from bitdust.dht import dht_service
    settings.init()
    lg.set_debug_level(30)
    dht_port = settings.getDHTPort()
    udp_port = settings.getUDPPort()
    if len(sys.argv) > 1:
        dht_port = int(sys.argv[1])
    if len(sys.argv) > 2:
        udp_port = int(sys.argv[2])
    dht_service.init(dht_port)
    dht_service.connect()
    udp.listen(udp_port)

    def _cb(result, typ, ip, details):
        print(result, typ, ip, details)
        A('shutdown')
        reactor.stop()  # @UndefinedVariable

    A('init', (udp_port))
    A('start', _cb)
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    from twisted.internet.defer import setDebugging
    setDebugging(True)
    # main()
    test_safe_stun()



#!/usr/bin/python
# service_customer_family.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_customer_family.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_customer_family
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return CustomerFamilyService()


class CustomerFamilyService(LocalService):

    service_name = 'service_customer_family'
    config_path = 'services/customer-family/enabled'

    def dependent_on(self):
        return [
            'service_supplier',
            'service_entangled_dht',
        ]

    def start(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.main import events
        from bitdust.contacts import contactsdb
        from bitdust.userid import id_url
        from bitdust.supplier import family_member
        from bitdust.transport import callback
        from bitdust.userid import my_id
        callback.append_inbox_callback(self._on_inbox_packet_received)
        for customer_idurl in contactsdb.customers():
            if not customer_idurl:
                continue
            if not id_url.is_cached(customer_idurl):
                continue
            if customer_idurl == my_id.getIDURL():
                lg.warn('skipping my own identity')
                continue
            fm = family_member.by_customer_idurl(customer_idurl)
            if not fm:
                fm = family_member.create_family(customer_idurl)
            fm.automat('init')
            local_customer_meta_info = contactsdb.get_customer_meta_info(customer_idurl)
            reactor.callLater(  # @UndefinedVariable
                0,
                fm.automat,
                'family-join',
                {
                    'supplier_idurl': my_id.getIDURL().to_bin(),
                    'ecc_map': local_customer_meta_info.get('ecc_map'),
                    'position': local_customer_meta_info.get('position', -1),
                    'family_snapshot': id_url.to_bin_list(local_customer_meta_info.get('family_snapshot')),
                },
            )
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        events.add_subscriber(self._on_existing_customer_accepted, 'existing-customer-accepted')
        events.add_subscriber(self._on_new_customer_accepted, 'new-customer-accepted')
        events.add_subscriber(self._on_existing_customer_terminated, 'existing-customer-terminated')
        return True

    def stop(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.main import events
        from bitdust.supplier import family_member
        events.remove_subscriber(self._on_new_customer_accepted, 'new-customer-accepted')
        events.remove_subscriber(self._on_existing_customer_accepted, 'existing-customer-accepted')
        events.remove_subscriber(self._on_existing_customer_terminated, 'existing-customer-terminated')
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        for fm in family_member.families().values():
            reactor.callLater(0, fm.automat, 'shutdown')  # @UndefinedVariable
        return True

    def _on_new_customer_accepted(self, evt):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.userid import my_id
        from bitdust.userid import id_url
        from bitdust.supplier import family_member
        customer_idurl = evt.data['idurl']
        fm = family_member.by_customer_idurl(customer_idurl)
        if not fm:
            fm = family_member.create_family(customer_idurl)
            fm.automat('init')
        else:
            lg.warn('family_member() instance already exists, but new customer just accepted %s' % customer_idurl)
        reactor.callLater(  # @UndefinedVariable
            0,
            fm.automat,
            'family-join',
            {
                'supplier_idurl': my_id.getIDURL().to_bin(),
                'ecc_map': evt.data.get('ecc_map'),
                'position': evt.data.get('position', -1),
                'family_snapshot': id_url.to_bin_list(evt.data.get('family_snapshot')),
            },
        )

    def _on_existing_customer_accepted(self, evt):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.supplier import family_member
        from bitdust.userid import id_url
        from bitdust.userid import my_id
        customer_idurl = evt.data['idurl']
        if customer_idurl == my_id.getIDURL():
            lg.warn('skipping my own identity')
            return
        if evt.data.get('position') is None:
            lg.warn('position of supplier in the family is still unclear')
            return
        fm = family_member.by_customer_idurl(customer_idurl)
        if not fm:
            lg.err('family_member() instance was not found for existing customer %s' % customer_idurl)
            return
        reactor.callLater(  # @UndefinedVariable
            0,
            fm.automat,
            'family-join',
            {
                'supplier_idurl': my_id.getIDURL().to_bin(),
                'ecc_map': evt.data.get('ecc_map'),
                'position': evt.data.get('position'),
                'family_snapshot': id_url.to_bin_list(evt.data.get('family_snapshot')),
            },
        )

    def _on_existing_customer_terminated(self, evt):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.supplier import family_member
        from bitdust.userid import my_id
        customer_idurl = evt.data['idurl']
        if customer_idurl == my_id.getIDURL():
            lg.warn('skipping my own identity')
            return
        fm = family_member.by_customer_idurl(customer_idurl)
        if not fm:
            lg.err('family_member() instance not found for existing customer %s' % customer_idurl)
            return
        reactor.callLater(  # @UndefinedVariable
            0,
            fm.automat,
            'family-leave',
            {
                'supplier_idurl': my_id.getIDURL().to_bin(),
                'ecc_map': evt.data.get('ecc_map'),
            },
        )

    def _on_incoming_contacts_packet(self, newpacket, info):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.lib import serialization
        from bitdust.lib import strng
        from bitdust.supplier import family_member
        from bitdust.userid import my_id
        from bitdust.userid import id_url
        try:
            json_payload = serialization.BytesToDict(newpacket.Payload, keys_to_text=True)
            contacts_type = strng.to_text(json_payload['type'])
            contacts_space = strng.to_text(json_payload['space'])
        except:
            lg.exc()
            return False

        if contacts_space != 'family_member':
            return False

        if contacts_type == 'suppliers_list':
            try:
                customer_idurl = id_url.field(json_payload['customer_idurl'])
                ecc_map = strng.to_text(json_payload['customer_ecc_map'])
                suppliers_list = id_url.fields_list(json_payload['suppliers_list'])
                transaction_revision = json_payload.get('transaction_revision')
            except:
                lg.exc()
                return False
            if customer_idurl.to_bin() == my_id.getIDURL().to_bin():
                lg.warn('received contacts for my own customer family')
                return False
            if not id_url.is_cached(customer_idurl):
                lg.warn('received contacts from unknown user: %r' % customer_idurl)
                return False
            fm = family_member.by_customer_idurl(customer_idurl)
            if not fm:
                lg.warn('family_member() instance not found for incoming %s from %s for customer %r' % (newpacket, info, customer_idurl))
                return False
            reactor.callLater(  # @UndefinedVariable
                0,
                fm.automat,
                'contacts-received',
                {
                    'type': contacts_type,
                    'packet': newpacket,
                    'customer_idurl': customer_idurl,
                    'customer_ecc_map': ecc_map,
                    'suppliers_list': suppliers_list,
                    'transaction_revision': transaction_revision,
                },
            )
            return True

        elif contacts_type == 'supplier_position':
            try:
                customer_idurl = id_url.field(json_payload['customer_idurl'])
                ecc_map = strng.to_text(json_payload['customer_ecc_map'])
                supplier_idurl = id_url.field(json_payload['supplier_idurl'])
                supplier_position = json_payload['supplier_position']
                family_snapshot = id_url.to_bin_list(json_payload.get('family_snapshot'))
            except:
                lg.exc()
                return False
            if customer_idurl.to_bin() == my_id.getIDURL().to_bin():
                lg.warn('received contacts for my own customer family')
                return False
            fm = family_member.by_customer_idurl(customer_idurl)
            if not fm:
                lg.warn('family_member() instance not found for incoming %s from %s for customer %r' % (newpacket, info, customer_idurl))
                return False
            reactor.callLater(  # @UndefinedVariable
                0,
                fm.automat,
                'contacts-received',
                {
                    'type': contacts_type,
                    'packet': newpacket,
                    'customer_idurl': customer_idurl,
                    'customer_ecc_map': ecc_map,
                    'supplier_idurl': supplier_idurl,
                    'supplier_position': supplier_position,
                    'family_snapshot': family_snapshot,
                },
            )
            return True

        return False

    def _on_inbox_packet_received(self, newpacket, info, *args):
        from bitdust.p2p import commands
        if newpacket.Command == commands.Contacts():
            return self._on_incoming_contacts_packet(newpacket, info)
        return False

    def _on_identity_url_changed(self, evt):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.userid import id_url
        from bitdust.supplier import family_member
        for customer_idurl, fm in family_member.families().items():
            if customer_idurl == id_url.field(evt.data['old_idurl']):
                customer_idurl.refresh(replace_original=True)
                fm.customer_idurl.refresh(replace_original=True)
                lg.info('found %r for customer with rotated identity and refreshed: %r' % (fm, customer_idurl))
                reactor.callLater(0, fm.automat, 'family-refresh')  # @UndefinedVariable
